from bokeh.layouts import column,row
import numpy as np
from bokeh.plotting import Figure, output_file, show
from bokeh.embed import components
from bokeh.models.widgets import Panel, Tabs
from bokeh.models import CustomJS, ColumnDataSource, Slider,Select
from bokeh.io import curdoc
from bokeh.layouts import row

def create_component_jwst(result_dict):
    """Generate front end plots JWST
    
    Function that is responsible for generating the front-end interactive plots for JWST.

    Parameters 
    ----------
    result_dict : dict 
        the dictionary returned from a PandExo run
    
    Returns
    -------
    tuple 
        A tuple containing `(script, div)`, where the `script` is the
        front-end javascript required, and `div` is a dictionary of plot
        objects.
    """  
    timing = result_dict['timing']
    noccultations = result_dict['timing']['Number of Transits']
    out = result_dict['PandeiaOutTrans']
    
    # select the tools we want
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"

    #Define units for x and y axis
    punit = result_dict['input']['Primary/Secondary']
    p=1.0
    if punit == 'fp/f*': p = -1.0
    else: punit = punit
    
    if result_dict['input']['Calculation Type'] =='phase_spec':
        x_axis_label='Time (secs)'
        frac = 1.0
    else:
        x_axis_label='Wavelength [microns]'
        frac = result_dict['timing']['Num Integrations Out of Transit']/result_dict['timing']['Num Integrations In Transit']

    electrons_out = result_dict['RawData']['electrons_out']
    electrons_in = result_dict['RawData']['electrons_in']
    
    var_in = result_dict['RawData']['var_in']
    var_out = result_dict['RawData']['var_out']
    
    
    x = result_dict['FinalSpectrum']['wave']
    y = result_dict['FinalSpectrum']['spectrum_w_rand']
    err = result_dict['FinalSpectrum']['error_w_floor']

    y_err = []
    x_err = []
    for px, py, yerr in zip(x, y, err):
        np.array(x_err.append((px, px)))
        np.array(y_err.append((py - yerr, py + yerr)))

    source = ColumnDataSource(data=dict(x=x, y=y, y_err=y_err, x_err=x_err, err=err, 
                                electrons_out=electrons_out, electrons_in=electrons_in, var_in=var_in, var_out=var_out, 
                                p=var_in*0+p,nocc=var_in*0+noccultations, frac = var_in*0+frac))
    original = ColumnDataSource(data=dict(x=x, y=y, y_err=y_err, x_err=x_err, err=err, electrons_out=electrons_out, electrons_in=electrons_in, var_in=var_in, var_out=var_out))

    ylims = [min(result_dict['OriginalInput']['model_spec'])- 0.1*min(result_dict['OriginalInput']['model_spec']),
                 0.1*max(result_dict['OriginalInput']['model_spec'])+max(result_dict['OriginalInput']['model_spec'])]
    xlims = [min(result_dict['FinalSpectrum']['wave']), max(result_dict['FinalSpectrum']['wave'])]

    plot_spectrum = Figure(plot_width=800, plot_height=300, x_range=xlims,
                               y_range=ylims, tools=TOOLS,#responsive=True,
                                 x_axis_label=x_axis_label,
                                 y_axis_label=punit, 
                               title="Original Model with Observation")
    
    plot_spectrum.line(result_dict['OriginalInput']['model_wave'],result_dict['OriginalInput']['model_spec'], color= "black", alpha = 0.5, line_width = 4)
        
    plot_spectrum.circle('x', 'y', source=source, line_width=3, line_alpha=0.6)
    plot_spectrum.multi_line('x_err', 'y_err', source=source)

    callback = CustomJS(args=dict(source=source, original=original), code="""
            // Grab some references to the data
            var sdata = source.get('data');
            var odata = original.get('data');

            // Create copies of the original data, store them as the source data
            sdata['x'] = odata['x'].slice(0);
            sdata['y'] = odata['y'].slice(0);

            sdata['y_err'] = odata['y_err'].slice(0);
            sdata['x_err'] = odata['x_err'].slice(0);
            sdata['err'] = odata['err'].slice(0);

            sdata['electrons_out'] = odata['electrons_out'].slice(0);
            sdata['electrons_in'] = odata['electrons_in'].slice(0);
            sdata['var_in'] = odata['var_in'].slice(0);
            sdata['var_out'] = odata['var_out'].slice(0);

            // Create some variables referencing the source data
            var x = sdata['x'];
            var y = sdata['y'];
            var y_err = sdata['y_err'];
            var x_err = sdata['x_err'];
            var err = sdata['err'];
            var p = sdata['p'];
            var frac = sdata['frac'];
            var og_ntran = sdata['nocc'];

            var electrons_out = sdata['electrons_out'];
            var electrons_in = sdata['electrons_in'];
            var var_in = sdata['var_in'];
            var var_out = sdata['var_out'];

            var wbin = wbin.get('value');
            var ntran = ntran.get('value');

            var ind = [];
            ind.push(0);
            var start = 1;


            for (i = 0; i < x.length-1; i++) {
                if (start == wbin) {
                    ind.push(i+1);
                    start =0;
                }
                start = start+1;
            }

            if (ind[ind.length-1] != x.length) {
                ind.push(x.length);
            }

            var xout = [];


            var foutout = [];
            var finout = [];
            var varinout = [];
            var varoutout = [];

            var xslice = []; 

            var foutslice = [];
            var finslice = [];
            var varoutslice = [];
            var varinslice = [];

            function add(a, b) {
                return a+b;
            }

            for (i = 0; i < ind.length-1; i++) {
                xslice = x.slice(ind[i],ind[i+1]);

                foutslice = electrons_out.slice(ind[i],ind[i+1]);
                finslice = electrons_in.slice(ind[i],ind[i+1]);
                
                varinslice = var_in.slice(ind[i],ind[i+1]);
                varoutslice = var_out.slice(ind[i],ind[i+1]);

                xout.push(xslice.reduce(add, 0)/xslice.length);
                foutout.push(foutslice.reduce(add, 0));
                finout.push(finslice.reduce(add, 0));
                
                varinout.push(varinslice.reduce(add, 0));
                varoutout.push(varoutslice.reduce(add, 0));

                xslice = [];
                foutslice = [];
                finslice = [];
                varinslice = [];
                varoutslice = [];
            }
            
            var new_err = 1.0;
            var rand = 1.0;

            for (i = 0; i < x.length; i++) {
                new_err = Math.pow((frac[i]/foutout[i]),2)*varinout[i] + Math.pow((finout[i]*frac[i]/Math.pow(foutout[i],2)),2)*varoutout[i];
                new_err = Math.sqrt(new_err)*Math.sqrt(og_ntran[i]/ntran);
                rand = new_err*(Math.random()-Math.random());
                y[i] = p[i]*((1.0 - frac[i]*finout[i]/foutout[i]) + rand); 
                x[i] = xout[i];
                x_err[i][0] = xout[i];
                x_err[i][1] = xout[i];
                y_err[i][0] = y[i] + new_err;
                y_err[i][1] = y[i] - new_err;            
            }

            source.trigger('change');
        """)

    #var_tot = (frac/electrons_out)**2.0 * var_in + (electrons_in*frac/electrons_out**2.0)**2.0 * var_out

    sliderWbin =  Slider(title="# pixels to bin", value=1, start=1, end=50, step= 1, callback=callback)
    callback.args["wbin"] = sliderWbin
    sliderTrans =  Slider(title="Num Trans", value=noccultations, start=1, end=50, step= 1, callback=callback)
    callback.args["ntran"] = sliderTrans
    layout = column(row(sliderWbin,sliderTrans), plot_spectrum)


    #out of transit 2d output 
    raw = result_dict['RawData']
    
    # Flux 1d
    x, y = raw['wave'], raw['e_rate_out']*result_dict['timing']['Seconds per Frame']*(timing["APT: Num Groups per Integration"]-1)
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]

    plot_flux_1d1 = Figure(tools=TOOLS,
                         x_axis_label='Wavelength [microns]',
                         y_axis_label='e-/integration', title="Flux Per Integration",
                         plot_width=800, plot_height=300)
    plot_flux_1d1.line(x, y, line_width = 4, alpha = .7)
    tab1 = Panel(child=plot_flux_1d1, title="Flux per Int")

    # BG 1d
    #x, y = out['1d']['extracted_bg_only']
    #y = y[~np.isnan(y)]
    #x = x[~np.isnan(y)]
    #plot_bg_1d1 = Figure(tools=TOOLS,
    #                     x_axis_label='Wavelength [microns]',
    #                     y_axis_label='Flux (e/s)', title="Background",
    #                     plot_width=800, plot_height=300)
    #plot_bg_1d1.line(x, y, line_width = 4, alpha = .7)
    #tab2 = Panel(child=plot_bg_1d1, title="Background Flux")

    # SNR 
    y = np.sqrt(y) #this is computing the SNR (sqrt of photons in a single integration)


    plot_snr_1d1 = Figure(tools=TOOLS,
                         x_axis_label=x_axis_label,
                         y_axis_label='sqrt(e-)/integration', title="SNR per integration",
                         plot_width=800, plot_height=300)
    plot_snr_1d1.line(x, y, line_width = 4, alpha = .7)
    tab3 = Panel(child=plot_snr_1d1, title="SNR per Int")


    # Error bars (ppm) 
    x = result_dict['FinalSpectrum']['wave']
    y = result_dict['FinalSpectrum']['error_w_floor']*1e6
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]    
    ymed = np.median(y)

    source2 = ColumnDataSource(data=dict(x=x, y=y,nocc=x*0+noccultations))
    original2 = ColumnDataSource(data=dict(x=x, y=y,nocc=x*0+noccultations))

    plot_noise_1d1 = Figure(tools=TOOLS,#responsive=True,
                         x_axis_label=x_axis_label,
                         y_axis_label='Spectral Precision (ppm)', title="Spectral Precision",
                         plot_width=800, plot_height=300, y_range = [0,2.0*ymed])
    ymed = np.median(y)
    plot_noise_1d1.circle('x', 'y', line_width = 4, alpha = .7, source=source2)

    callback2 = CustomJS(args=dict(source=source2, original=original2), code="""
            // Grab some references to the data
            var sdata = source.get('data');
            var odata = original.get('data');

            // Create copies of the original data, store them as the source data
            sdata['x'] = odata['x'].slice(0);
            sdata['y'] = odata['y'].slice(0);

            // Create some variables referencing the source data
            var x = sdata['x'];
            var y = sdata['y'];

            var og_ntran = sdata['nocc'];

            var wbin = wbin.get('value');
            var ntran = ntran.get('value');

            var ind = [];
            ind.push(0);
            var start = 1;


            for (i = 0; i < x.length-1; i++) {
                if (start == wbin) {
                    ind.push(i+1);
                    start =0;
                }
                start = start+1;
            }

            if (ind[ind.length-1] != x.length) {
                ind.push(x.length);
            }

            var xout = [];
            var yout = [];

            var xslice = []; 
            var yslice = [];

            function addpow(tot, num) {
                return tot+ Math.pow(num,2);
            }
            function add(a, b) {
                return a+b;
            }
            for (i = 0; i < ind.length-1; i++) {

                xslice = x.slice(ind[i],ind[i+1]);
                yslice = y.slice(ind[i],ind[i+1]);

                xout.push(xslice.reduce(add, 0)/xslice.length);
                yout.push(Math.sqrt(yslice.reduce(addpow, 0))/yslice.length);

                xslice = [];
                yslice = [];
            }
            
            var new_err = 1.0;
            for (i = 0; i < x.length; i++) {
                new_err = yout[i]*Math.sqrt(og_ntran[i]/ntran);
                x[i] = xout[i];
                y[i] = new_err       
            }

            source.trigger('change');
        """)    

    sliderWbin2 =  Slider(title="# pixels to bin", value=1, start=1, end=50, step= 1, callback=callback2)
    callback2.args["wbin"] = sliderWbin2   
    sliderTrans2 =  Slider(title="Num Trans", value=noccultations, start=1, end=50, step= 1, callback=callback2)
    callback2.args["ntran"] = sliderTrans2
    noise_lay = column(row(sliderWbin2, sliderTrans2) , plot_noise_1d1)
    tab4 = Panel(child=noise_lay, title="Precision")

    #Not happy? Need help picking a different mode? 
    plot_spectrum2 = Figure(plot_width=800, plot_height=300, x_range=xlims,y_range=ylims, tools=TOOLS,
                             x_axis_label=x_axis_label,
                             y_axis_label=punit, title="Original Model",y_axis_type="log")

    plot_spectrum2.line(result_dict['OriginalInput']['model_wave'],result_dict['OriginalInput']['model_spec'],
                        line_width = 4,alpha = .7)
    tab5 = Panel(child=plot_spectrum2, title="Original Model")


    #create set of five tabs 
    tabs1d = Tabs(tabs=[ tab1,tab3, tab4, tab5])



    # Detector 2d
    data = out['2d']['detector']

    
    xr, yr = data.shape
    
    plot_detector_2d = Figure(tools="pan,wheel_zoom,box_zoom,reset,hover,save",
                         x_range=[0, yr], y_range=[0, xr],
                         x_axis_label='Pixel', y_axis_label='Spatial',
                         title="2D Detector Image",
                        plot_width=800, plot_height=300)
    
    plot_detector_2d.image(image=[data], x=[0], y=[0], dh=[xr], dw=[yr],
                      palette="Spectral11")


    #2d tabs 

    #2d snr 
    data = out['2d']['snr']
    data[np.isinf(data)] = 0.0
    xr, yr = data.shape
    plot_snr_2d = Figure(tools=TOOLS,
                         x_range=[0, yr], y_range=[0, xr],
                         x_axis_label='Pixel', y_axis_label='Spatial',
                         title="Signal-to-Noise Ratio",
                        plot_width=800, plot_height=300)
    
    plot_snr_2d.image(image=[data], x=[0], y=[0], dh=[xr], dw=[yr],
                      palette="Spectral11")
    
    tab1b = Panel(child=plot_snr_2d, title="SNR")

    #saturation
    
    data = out['2d']['saturation']
    xr, yr = data.shape
    plot_sat_2d = Figure(tools=TOOLS,
                         x_range=[0, yr], y_range=[0, xr],
                         x_axis_label='Pixel', y_axis_label='Spatial',
                         title="Saturation",
                        plot_width=800, plot_height=300)
    
    plot_sat_2d.image(image=[data], x=[0], y=[0], dh=[xr], dw=[yr],
                      palette="Spectral11")
    
    tab2b = Panel(child=plot_sat_2d, title="Saturation")

    tabs2d = Tabs(tabs=[ tab1b, tab2b])
    
 
    result_comp = components({'plot_spectrum':layout, 
                              'tabs1d': tabs1d, 'det_2d': plot_detector_2d,
                              'tabs2d': tabs2d})

    return result_comp
    


def create_component_hst(result_dict):
    """Generate front end plots HST
    
    Function that is responsible for generating the front-end spectra plots for HST.
    
    Parameters
    ----------
    result_dict : dict 
        The dictionary returned from a PandExo (HST) run

    Returns
    -------
    tuple
        A tuple containing `(script, div)`, where the `script` is the
        front-end javascript required, and `div` is a dictionary of plot
        objects.
    """                                   
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"

    #plot planet spectrum
    mwave = result_dict['planet_spec']['model_wave']
    mspec = result_dict['planet_spec']['model_spec']
    
    binwave = result_dict['planet_spec']['binwave']
    binspec = result_dict['planet_spec']['binspec']
    
    error = result_dict['planet_spec']['error']
    error = np.zeros(len(binspec))+ error
    xlims = [result_dict['planet_spec']['wmin'], result_dict['planet_spec']['wmax']]
    ylims = [np.min(binspec)-2.0*error[0], np.max(binspec)+2.0*error[0]]
    
    eventType = result_dict['calc_start_window']['eventType'] 

    if eventType=='transit':
        y_axis = '(Rp/R*)^2'
    elif eventType =='eclipse':
        y_axis='Fp/F*'

    plot_spectrum = Figure(plot_width=800, plot_height=300, x_range=xlims,
                               y_range=ylims, tools=TOOLS,#responsive=True,
                                 x_axis_label='Wavelength [microns]',
                                 y_axis_label=y_axis, 
                               title="Original Model with Observation")
    
    y_err = []
    x_err = []
    for px, py, yerr in zip(binwave, binspec, error):
        np.array(x_err.append((px, px)))
        np.array(y_err.append((py - yerr, py + yerr)))

    plot_spectrum.line(mwave,mspec, color= "black", alpha = 0.5, line_width = 4)
    plot_spectrum.circle(binwave,binspec, line_width=3, line_alpha=0.6)
    plot_spectrum.multi_line(x_err, y_err)
    
    
    #earliest and latest start times without ramp
    obsphase1 = result_dict['calc_start_window']['obsphase1']
    obstr1 = result_dict['calc_start_window']['obstr1']
    rms = result_dict['calc_start_window']['light_curve_rms']
    obsphase2 = result_dict['calc_start_window']['obsphase2']
    obstr2 = result_dict['calc_start_window']['obstr2']
    phase1 = result_dict['calc_start_window']['phase1']    
    phase2 = result_dict['calc_start_window']['phase2']
    trmodel1 = result_dict['calc_start_window']['trmodel1']
    trmodel2 = result_dict['calc_start_window']['trmodel2']    
    
    if isinstance(rms, float):
        rms = np.zeros(len(obsphase1))+rms
    y_err1 = []
    x_err1 = []
    for px, py, yerr in zip(obsphase1, obstr1, rms):
        np.array(x_err1.append((px, px)))
        np.array(y_err1.append((py - yerr, py + yerr)))

    y_err2 = []
    x_err2 = []
    for px, py, yerr in zip(obsphase2, obstr2, rms):
        np.array(x_err2.append((px, px)))
        np.array(y_err2.append((py - yerr, py + yerr)))

    early = Figure(plot_width=400, plot_height=300,
                               tools=TOOLS,#responsive=True,
                                 x_axis_label='Orbital Phase',
                                 y_axis_label='Flux', 
                               title="Earliest Start Time")
    
    early.line(phase1, trmodel1, color='black',alpha=0.5, line_width = 4)
    early.circle(obsphase1, obstr1, line_width=3, line_alpha=0.6)
    early.multi_line(x_err1, y_err1)
     
    late = Figure(plot_width=400, plot_height=300, 
                                tools=TOOLS,#responsive=True,
                                 x_axis_label='Orbital Phase',
                                 y_axis_label='Flux', 
                               title="Latest Start Time")
    late.line(phase2, trmodel2, color='black',alpha=0.5, line_width = 3)
    late.circle(obsphase2, obstr2, line_width=3, line_alpha=0.6)
    late.multi_line(x_err2, y_err2)
        
    start_time = row(early, late)
    tab1b = Panel(child=start_time, title="Timing")
    
    #now create plot that has optional ramp
    obsphase1 = result_dict['light_curve']['obsphase1']
    rms = result_dict['light_curve']['light_curve_rms']
    obsphase2 = result_dict['light_curve']['obsphase2']
    phase1 = result_dict['light_curve']['phase1']
    phase2 = result_dict['light_curve']['phase2']
    counts1 = result_dict['light_curve']['counts1']
    counts2 = result_dict['light_curve']['counts2']
    count_noise = result_dict['light_curve']['count_noise']
    ramp_included = result_dict['light_curve']['ramp_included']
    model_counts1 = result_dict['light_curve']['model_counts1']
    model_counts2 = result_dict['light_curve']['model_counts2']

    if isinstance(count_noise, float):
        rms = np.zeros(len(counts1)) + count_noise
    y_err1 = []
    x_err1 = []
    for px, py, yerr in zip(obsphase1, counts1, rms):
        np.array(x_err1.append((px, px)))
        np.array(y_err1.append((py - yerr, py + yerr)))

    y_err2 = []
    x_err2 = []
    for px, py, yerr in zip(obsphase2, counts2, rms):
        np.array(x_err2.append((px, px)))
        np.array(y_err2.append((py - yerr, py + yerr)))

    if ramp_included:
        title_description = " (Ramp Included)"
    else:
        title_description =" (Ramp Removed)"

    early = Figure(plot_width=400, plot_height=300,
                               tools=TOOLS,#responsive=True,
                                 x_axis_label='Orbital Phase',
                                 y_axis_label='Flux [electrons/pixel]',
                               title="Earliest Start Time" + title_description)


    early.line(phase1, model_counts1, color='black', alpha=0.5, line_width=4)
    early.circle(obsphase1, counts1, line_width=3, line_alpha=0.6)
    early.multi_line(x_err1, y_err1)

    late = Figure(plot_width=400, plot_height=300,
                  tools=TOOLS,  # responsive=True,
                  x_axis_label='Orbital Phase',
                  y_axis_label='Flux [electrons/pixel]',
                  title="Latest Start Time" + title_description)

    late.line(phase2, model_counts2, color='black', alpha=0.5, line_width=3)
    late.circle(obsphase2, counts2, line_width=3, line_alpha=0.6)
    late.multi_line(x_err2, y_err2)

    start_time = row(early, late)


    tab2b = Panel(child=start_time, title="Electrons")

    tabs2d = Tabs(tabs=[ tab1b, tab2b])
    
    
    result_comp = components({'plot_spectrum':plot_spectrum, 
                              'start_time':tabs2d})

    return result_comp


 



def wrapper(dictinput, verbose=False):
    """Tpo level function to call either jwst, hst, wfirst
    
    Top level function which calls either jwst, hst or wfirst noise simulation. 
    
    Parameters
    ----------
    dictinput : 
        dictionary containing instrument parameters and exoplanet 
        specific parameters. {"pandeia_input":dict1, "pandexo_input":dict1}
    verbose : bool 
        (Optional) prints out checkpoints throughout code

    Returns
    -------
    dict 
        output specific to observatory requested

    
    Notes
    -----
    You should not run simulations through this. It is much easier to run simulations through 
    either **run_online** or **justdoit**. **justdoit** contains functions 
    to create input dictionaries and **run_online** contains web forms to create input dictionaries.
    
    See Also
    --------
    pandexo.engine.justdoit : gives ability to simply submit runs 
    pandexo.engine.run_online : submit runs through user interface 
    """
    pandeia_input = dictinput['pandeia_input']    	
    telescope = pandeia_input['telescope']

    if telescope=='jwst':
        from .jwst import compute_full_sim
        return compute_full_sim(dictinput, verbose=verbose)
    elif telescope=='hst':
        from .hst import compute_sim_hst
        return compute_sim_hst(dictinput, verbose=verbose)
    elif telescope=='wfirst':
        return
    else:
        print("INVALID TELESCOPE. PandExo only accepts: jwst, hst, wfirst")



# This is a special __init__.py required to for namespace packages.
# There should be no other code in this module.
#try:
#    from pkg_resources import declare_namespace
#    declare_namespace(__name__)
#except ImportError:
#    from pkgutil import extend_path
#    __path__ = extend_path(__path__, __name__)




#!/usr/bin/env python

# This sample setup.py can be used as a template for any project using d2to1.
# Simply copy this file and, if desired, delete all the comments.  Also remove
# the 'namespace_packages' and 'packages' arguments to setup.py if this project
# does not contain any packages beloning to a namespace package.

# This import statement attempts to import the setup() function from setuptools
# (this replaces the setup() one uses from distutils.core when using plain
# distutils).
#
# If the import fails (the user doesn't have the setuptools package) it then
# uses the ez_setup bootstrap script to install setuptools, then retries the
# import.  This is common practice for packages using setuptools.
try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup
import sys 
PY_V = sys.version_info
# The standard setup() call.  Notice, however, that most of the arguments
# normally passed to setup() are absent.  They will instead be read from the
# setup.cfg file using d2to1.
#
# In order for this to work it is necessary to specify setup_requires=['d2to1']
# If the user does not have d2to1, this will boostrap it.  Also require
# stsci.distutils to use any of the common setup_hooks included in
# stsci.distutils (see the setup.cfg for more details).
#
# The next line, which defines namespace_packages and packages is only
# necessary if this projet contains a package belonging to the stsci namespace
# package, such as stsci.distutils or stsci.tools.  This is not necessary for
# projects with their own namespace, such as acstools or pyraf.
#
# d2to1=True is required to enable d2to1 and read the remaning project metadata
# from the setup.cfg file.
#
# use_2to3 and zip_safe are common options support by setuptools; these can
# also be placed in the setup.cfg, as will be demonstrated in a future update
# to this sample package.

if sys.version_info < (3,0): 
    pandas_version = '0.24.0'
else:
    pandas_version = '0.25.0'
setup(

    name='pandexo.engine',
    version='1.5.3',
    summary='pandexo transiting exoplanet simulator',
    description_file='README.rst',
    author='Natasha Batalha at NASA Ames',
    author_email='natasha.e.batalha@nasa.gov',
    home_page='https://natashabatalha.github.io/PandExo',
    license='GPL',
    url='https://github.com/natashabatalha/PandExo',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    packages=[
        'pandexo',
        'pandexo.engine',
        'pandexo.engine.reference',
        'pandexo.engine.static',
        'pandexo.engine.static.css',
        'pandexo.engine.static.fonts',
        'pandexo.engine.static.img',
        'pandexo.engine.static.js',
        'pandexo.engine.temp',
        'pandexo.engine.templates',
        'pandexo.engine.utils'
    ],
    package_dir={
        'engine': 'pandexo/engine'
    },
    include_package_data=True,
    # package_data = {
    #            'engine' : ['templates','*.html'],
    #            'engine': ['reference','*.json'],
    #            'engine': ['static','css','*'],
    #            'engine': ['static','fonts','*'],
    #            'engine': ['static','img','*'],
    #            'engine': ['static','js','*']
    # },

    install_requires=[
          'numpy',
          'bokeh==0.12.6',
          'tornado',
          'pandas=='+pandas_version,
          'joblib',
          'pandeia.engine==1.5.1',
          'batman-package',
          'photutils',
          'astropy',
          'pysynphot',
          'sqlalchemy',
          'astroquery',
          'scipy==1.3.1'
          ],
    entry_points = {
        'console_scripts':
            ['start_pandexo=pandexo.engine.run_online:main']
    },
    zip_safe=False,

)



import os
import sys
import json
import numpy as np
import pandas as pd
from copy import deepcopy 
from astropy.io import fits
from pandeia.engine.instrument_factory import InstrumentFactory
from pandeia.engine.perform_calculation import perform_calculation
from . import create_input as create
from .compute_noise import ExtractSpec
import astropy.units as u
import pickle
from pandeia.engine.calc_utils import build_default_calc, build_default_source

#constant parameters.. consider putting these into json file 
#max groups in integration
max_ngroup = 65536.0 
#minimum number of integrations
min_nint_trans = 1

#refdata directory
default_refdata_directory = os.environ.get("pandeia_refdata")

def compute_full_sim(dictinput,verbose=False): 
    """Top level function to set up exoplanet obs. for JW
    
    Function to set up explanet observations for JWST only and 
    compute simulated spectrum. It uses STScI's Pandeia to compute 
    instrument throughputs and WebbPSF to compute PSFs. 
    
    Parameters
    ----------
    dictinput : dict
        dictionary containing instrument parameters and exoplanet specific 
        parameters. {"pandeia_input":dict1, "pandexo_input":dict1}
    verbose : bool 
        (Optional) prints out check points throughout code 
    
    Returns
    -------
    dict
        large dictionary with 1d, 2d simualtions, timing info, instrument info, warnings
    
    Examples
    --------  
      
    >>> from .pandexo.engine.jwst import compute_full_sim 
    >>> from .pandexo.engine.justplotit import jwst_1d_spec
    >>> a = compute_full_sim({"pandeia_input": pandeiadict, "pandexo_input":exodict})
    >>> jwst_1d_spec(a)
    .. image:: 1d_spec.png
    
    Notes
    -----
    It is much easier to run simulations through either **run_online** or **justdoit**. **justdoit** contains functions to create input dictionaries and **run_online** contains web forms to create input dictionaries.
    
    See Also
    -------- 
        pandexo.engine.justdoit.run_pandexo : Best function for running pandexo runs
        pandexo.engine.run_online : Allows running functions through online interface
    """
    pandeia_input = dictinput['pandeia_input']
    pandexo_input = dictinput['pandexo_input']    	
	
    #define the calculation we'll be doing 
    if pandexo_input['planet']['w_unit'] == 'sec':
        calculation = 'phase_spec'
    else: 
        calculation = pandexo_input['calculation'].lower()

    #which instrument 
    instrument = pandeia_input['configuration']['instrument']['instrument']
    conf = pandeia_input['configuration']
    
    #if optimize is in the ngroups section, this will throw an error 
    #so create temp conf with 2 groups 
    try:
        i = InstrumentFactory(config=conf)
        det_pars = i.get_detector_pars()
        exp_pars = i.get_exposure_pars()
    except: 
        conf_temp = deepcopy(conf) 
        conf_temp['detector']['ngroup'] = 2 
        i = InstrumentFactory(config=conf_temp)
        det_pars = i.get_detector_pars()
        exp_pars = i.get_exposure_pars()
        
    #detector parameters
    fullwell = det_pars['fullwell'] #from pandeia data
    rn = det_pars['rn']
    sat_unit = pandexo_input['observation']['sat_unit']

    if sat_unit =='%':
        sat_level = pandexo_input['observation']['sat_level']/100.0*fullwell
    elif sat_unit =='e':
        sat_level = pandexo_input['observation']['sat_level']
    else: 
        raise Exception("Saturation Level Needs Units: % fullwell or Electrons ")

    mingroups = det_pars['mingroups']
    
    #exposure parameters 
    tframe = exp_pars.tframe
    nframe = exp_pars.nframe
    nskip = exp_pars.nskip
    
    #parameteres needed from exo_input
    mag = pandexo_input['star']['mag']
    
    
    noccultations = pandexo_input['observation']['noccultations']
    R = pandexo_input['observation']['R']

    noise_floor = pandexo_input['observation']['noise_floor']

    
    #get stellar spectrum and in transit spec
    star_spec = create.outTrans(pandexo_input['star'])
    #get rstar if user calling from grid 
    both_spec = create.bothTrans(star_spec, pandexo_input['planet'], star=pandexo_input['star'])
    out_spectrum = np.array([both_spec['wave'], both_spec['flux_out_trans']])
    
    #get transit duration from phase curve or from input 
    if calculation == 'phase_spec': 
        transit_duration = max(both_spec['time']) - min(both_spec['time'])
    else: 
        #convert to seconds, then remove quantity and convert back to float 
        transit_duration = float((pandexo_input['planet']['transit_duration']*u.Unit(pandexo_input['planet']['td_unit'])).to(u.second)/u.second)

    #amount of exposure time out-of-occultation, as a fraction of in-occ time 
    try:
        expfact_out = pandexo_input['observation']['fraction'] 
        print("WARNING: key input fraction has been replaced with new 'baseline option'. See notebook example")
        pandexo_input['observation']['baseline'] = pandexo_input['observation']['fraction'] 
        pandexo_input['observation']['baseline_unit'] ='frac'
    except:
        if pandexo_input['observation']['baseline_unit'] =='frac':
            expfact_out = pandexo_input['observation']['baseline'] 
        elif pandexo_input['observation']['baseline_unit'] =='total':
            expfact_out = transit_duration/( pandexo_input['observation']['baseline'] - transit_duration)
        elif pandexo_input['observation']['baseline_unit'] =='total_hrs':
            expfact_out = transit_duration/( pandexo_input['observation']['baseline']*3600.0 - transit_duration)
        else: 
            raise Exception("Wrong units for baseine: either 'frac' or 'total' or 'total_hrs' accepted")

    #add to pandeia input 
    pandeia_input['scene'][0]['spectrum']['sed']['spectrum'] = out_spectrum
    
    if isinstance(pandeia_input["configuration"]["detector"]["ngroup"], (float,int)):
        m = {"ngroup":pandeia_input["configuration"]["detector"]["ngroup"], "tframe":tframe,
            "nframe":nframe,"mingroups":mingroups,"nskip":nskip}
    else:
        #run pandeia once to determine max exposure time per int and get exposure params
        if verbose: print("Optimization Reqested: Computing Duty Cycle")
        m = {"maxexptime_per_int":compute_maxexptime_per_int(pandeia_input, sat_level) , 
            "tframe":tframe,"nframe":nframe,"mingroups":mingroups,"nskip":nskip}
        if verbose: print("Finished Duty Cycle Calc")

    #calculate all timing info
    timing, flags = compute_timing(m,transit_duration,expfact_out,noccultations)
    
    #Simulate out trans and in transit
    if verbose: print("Starting Out of Transit Simulation")
    out = perform_out(pandeia_input, pandexo_input,timing, both_spec)
    
    #extract extraction area before dict conversion
    extraction_area = out.extraction_area
    out = out.as_dict()
    out.pop('3d')
    if verbose: print("End out of Transit")

    #Remove effects of Quantum Yield from shot noise 
    out = remove_QY(out, instrument)

    #this kind of redundant going to compute inn from out instead 
    #keep perform_in but change inputs to (out, timing, both_spec)
    if verbose: print("Starting In Transit Simulation")
    inn = perform_in(pandeia_input, pandexo_input,timing, both_spec, out, calculation)
    if verbose: print("End In Transit")
    


    #compute warning flags for timing info 
    warnings = add_warnings(out, timing, sat_level/fullwell, flags, instrument) 

    compNoise = ExtractSpec(inn, out, rn, extraction_area, timing)
    
    #slope method is pandeia's pure noise calculation (taken from SNR)
    #contains correlated noise, RN, dark current, sky, 
    #uses MULTIACCUM formula so we deviated from this. 
    #could eventually come back to this if Pandeia adopts First-Last formula
    if calculation == 'slope method': 
        #Extract relevant info from pandeia output (1d curves and wavelength) 
        #extracted flux in units of electron/s
        w = out['1d']['extracted_flux'][0]
        result = compNoise.run_slope_method()

    #derives noise from 2d postage stamps. Doing this results in a higher 
    #1d flux rate than the Pandeia gets from extracting its own. 
    #this should be used to benchmark Pandeia's 1d extraction  
    elif calculation == '2d extract':
        w = out['1d']['extracted_flux'][0]
        result = compNoise.run_2d_extract()
    
    #this is the noise calculation that PandExo uses online. It derives 
    #its own calculation of readnoise and does not use MULTIACUMM 
    #noise formula  
    elif calculation == 'fml':
        w = out['1d']['extracted_flux'][0]
        result = compNoise.run_f_minus_l()
    
    elif calculation == 'phase_spec':
        result = compNoise.run_phase_spec()
        w = result['time']
    else:
        result = None
        raise Exception('WARNING: Calculation method not found.')
        
    varin = result['var_in_1d']
    varout = result['var_out_1d']
    extracted_flux_out = result['photon_out_1d']
    extracted_flux_inn = result['photon_in_1d']

        
    #bin the data according to user input 
    if R != None: 
        wbin = bin_wave_to_R(w, R)

        photon_out_bin = uniform_tophat_sum(wbin, w,extracted_flux_out)
        photon_in_bin = uniform_tophat_sum(wbin,w, extracted_flux_inn)
        var_in_bin = uniform_tophat_sum(wbin, w,varin)
        var_out_bin = uniform_tophat_sum(wbin,w, varout)

        wbin = wbin[photon_out_bin > 0 ]
        photon_in_bin = photon_in_bin[photon_out_bin > 0 ]
        var_in_bin = var_in_bin[photon_out_bin > 0 ]
        var_out_bin = var_out_bin[photon_out_bin > 0 ]
        photon_out_bin = photon_out_bin[photon_out_bin>0]
    else: 
        wbin = w
        photon_out_bin = extracted_flux_out
        wbin = wbin[photon_out_bin>0]
        photon_in_bin = extracted_flux_inn
        photon_in_bin = photon_in_bin[photon_out_bin>0]
        var_in_bin = varin
        var_in_bin = var_in_bin[photon_out_bin>0]
        var_out_bin = varout
        var_out_bin = var_out_bin[photon_out_bin>0]
        photon_out_bin = photon_out_bin[photon_out_bin>0]
        
    
    if calculation == 'phase_spec':
        to = (timing["APT: Num Groups per Integration"]-1)*tframe
        ti = (timing["APT: Num Groups per Integration"]-1)*tframe
    else: 
        #otherwise error propagation and account for different 
        #times in transit and out 
        to = result['on_source_out']
        ti = result['on_source_in']
        
    var_tot = (to/ti/photon_out_bin)**2.0 * var_in_bin + (photon_in_bin*to/ti/photon_out_bin**2.0)**2.0 * var_out_bin
    error_spec = np.sqrt(var_tot)
        
    #factor in occultations to noise 
    nocc = timing['Number of Transits']
    error_spec = error_spec / np.sqrt(nocc)
        
    #Add in user specified noise floor 
    error_spec_nfloor = add_noise_floor(noise_floor, wbin, error_spec) 

    #add in random noise for the simulated spectrum 
    np.random.seed()
    rand_noise= error_spec_nfloor*(np.random.randn(len(wbin)))
    raw_spec = (photon_out_bin/to-photon_in_bin/ti)/(photon_out_bin/to)
    sim_spec = raw_spec + rand_noise 
    
    #if secondary tranist, multiply spectra by -1 
    if pandexo_input['planet']['f_unit'] == 'fp/f*':
        sim_spec = -1.0*sim_spec
        raw_spec = -1.0*raw_spec    
   
    #package processed data
    finalspec = {'wave':wbin,
              'spectrum': raw_spec,
              'spectrum_w_rand':sim_spec,
              'error_w_floor':error_spec_nfloor}
    
    rawstuff = {
                'electrons_out':photon_out_bin*nocc, 
                'electrons_in':photon_in_bin*nocc,
                'var_in':var_in_bin*nocc, 
                'var_out':var_out_bin*nocc,
                'e_rate_out':photon_out_bin/to,
                'e_rate_in':photon_in_bin/ti,
                'wave':wbin,
                'error_no_floor':error_spec, 
                'rn[out,in]':result['rn[out,in]'],
                'bkg[out,in]':result['bkg[out,in]']
                }
 
    result_dict = as_dict(out,both_spec ,finalspec, 
                timing, mag, sat_level, warnings,
                pandexo_input['planet']['f_unit'], rawstuff,calculation)

    return result_dict 
    
def compute_maxexptime_per_int(pandeia_input, sat_level):
    """Computes optimal maximum exposure time per integration
    
    Function to simulate 2d jwst image with 2 groups, 1 integration, 1 exposure 
    and return the maximum time 
    for one integration before saturation occurs. If saturation has 
    already occured, returns maxexptime_per_int as np.nan. This then 
    tells Pandexo to set min number of groups (ngroups =2). This avoids 
    error if saturation occurs. This routine assumes that min ngroups is 2. 
    
    Parameters
    ----------
    pandeia_input : dict 
        pandeia dictionary input 
    sat_level : int or float
        user defined saturation level in units of electrons
    
    Returns
    ------- 
    float 
        Maximum exposure time per integration before specified saturation level
    
    Examples
    --------
    
    >>> max_time = compute_maxexptime_per_int(pandeia_input, 50000.0)
    >>> print(max_time)
    12.0
    """
    
    #run once to get 2d rate image 
    pandeia_input['configuration']['detector']['ngroup'] = 2 
    pandeia_input['configuration']['detector']['nint'] = 1 
    pandeia_input['configuration']['detector']['nexp'] = 1
    
    report = perform_calculation(pandeia_input, dict_report=False)
    report_dict = report.as_dict() 
    
    # count rate on the detector in e-/second/pixel
    #det = report_dict['2d']['detector']
    det = report.signal.rate_plus_bg_list[0]['fp_pix']

    timeinfo = report_dict['information']['exposure_specification']
    #totaltime = timeinfo['tgroup']*timeinfo['ngroup']*timeinfo['nint']
    
    maxdetvalue = np.max(det)


    #maximum time before saturation per integration 
    #based on user specified saturation level

    try:
        maxexptime_per_int = sat_level/maxdetvalue
    except: 
        maxexptime_per_int = np.nan
    
    return maxexptime_per_int
        
def compute_timing(m,transit_duration,expfact_out,noccultations): 
    """Computes all timing info for observation
    
    Computes all JWST specific timing info for observation including. Some pertinent 
    JWST terminology:

        - frame: The result of sequentially clocking and digitizing all pixels in a rectangular area of an SCA. **Full-fame readout** means to digitize all pixels in an SCA, including reference pixels. **Frame** also applies to the result of clocking and digitizing a subarray on an SCA.
        - group: One or more consecutively read frames. There are no intervening resets. Frames may be averaged to form a group but for exoplanets the read out scheme is always 1 frame = 1 group
        - integration: The end result of resetting the detector and then non-destructively sampling it one or more times over a finite period of time before resetting the detector again. This is a unit of data for which signal is proportional to intensity, and it consists of one or more GROUPS.
        - exposure: The end result of one or more INTEGRATIONS over a finite period of time.  EXPOSURE defines the contents of a single FITS file.
    
    Parameters
    ---------
    m : dict 
        Dictionary output from **compute_maxexptime_per_int**
    transit_duration : float or int 
        transit duration in seconds 
    expfact_out : float or int 
        fraction of time spent in transit versus out of transit 
    noccultations : int 
        number of transits 
    
    Returns
    ------- 
    timing : dict
        All timing info
    warningflag : dict 
        Warning flags
    
    Examples
    --------
    >>> timing, flags = compute_timing(m, 2*60.0*60.0, 1.0, 1.0)
    >>> print((list(timing.keys())))
    ['Number of Transits', 'Num Integrations Out of Transit', 'Num Integrations In Transit', 
    'APT: Num Groups per Integration', 'Seconds per Frame', 'Observing Efficiency (%)', 'On Source Time(sec)', 
    'Exposure Time Per Integration (secs)', 'Reset time Plus 30 min TA time (hrs)',
    'APT: Num Integrations per Occultation', 'Transit Duration']
    """
    tframe = m['tframe']
    nframe = m['nframe']
    nskip = m['nskip']
    mingroups = m['mingroups']
    overhead_per_int = tframe #overhead time added per integration 
    try: 
        #are we starting with a exposure time ?
        maxexptime_per_int = m['maxexptime_per_int']
    except:
        #or a pre defined number of groups specified by user
        ngroups_per_int = m['ngroup']
        
    flag_default = "All good"
    flag_high = "All good"
    if 'maxexptime_per_int' in locals():
        #Frist, if maxexptime_per_int has been defined (from above), compute ngroups_per_int
        
        #number of frames in one integration is the maximum time beofre exposure 
        #divided by the time it takes for one frame. Note this does not include 
        #reset frames 

        nframes_per_int = np.floor(maxexptime_per_int/tframe)
    
        #for exoplanets nframe =1 an nskip always = 0 so ngroups_per_int 
        #and nframes_per_int area always the same 
        ngroups_per_int = np.floor(nframes_per_int/(nframe + nskip)) 
    
        #put restriction on number of groups 
        #there is a hard limit to the maximum number groups. 
        #if you exceed that limit, set it to the maximum value instead.
        #also set another check for saturation

        if ngroups_per_int > max_ngroup:
            ngroups_per_int = max_ngroup
            flag_high = "Groups/int > max num of allowed groups"
 
        if (ngroups_per_int < mingroups) | np.isnan(ngroups_per_int):
            ngroups_per_int = mingroups  
            nframes_per_int = mingroups
            flag_default = "NGROUPS<"+str(mingroups)+"SET TO NGROUPS="+str(mingroups)

    elif 'ngroups_per_int' in locals(): 
        #if it maxexptime_per_int been defined then set nframes per int 
        nframes_per_int = ngroups_per_int*(nframe+nskip)
        
        #if that didn't work its because maxexptime_per_int is nan .. run calc with mingroups
    else:
        #if maxexptime_per_int is nan then just ngroups and nframe to 2 
        #for the sake of not returning error
        ngroups_per_int = mingroups
        nframes_per_int = mingroups
        flag_default = "Something went wrong. SET TO NGROUPS="+str(mingroups)

          
    #the integration time is related to the number of groups and the time of each 
    #group 
    exptime_per_int = ngroups_per_int*tframe
    
    #clock time includes the reset frame 
    clocktime_per_int = (ngroups_per_int+1.0)*tframe
    
    #observing efficiency (i.e. what percentage of total time is spent on soure)
    eff = (ngroups_per_int - 1.0)/(ngroups_per_int + 1.0)
    
    #this says "per occultation" but this is just the in transit frames.. See below
    # transit duration / ((ngroups + reset)*frame time)
    nint_per_occultation =  transit_duration/((ngroups_per_int+1.0)*tframe)
    
    #figure out how many integrations are in transit and how many are out of transit 
    nint_in = np.ceil(nint_per_occultation)
    nint_out = np.ceil(nint_in/expfact_out)
    
    #you would never want a single integration in transit. 
    #here we assume that for very dim things, you would want at least 
    #3 integrations in transit 
    if nint_in < min_nint_trans:
        ngroups_per_int = np.floor(ngroups_per_int/min_nint_trans)
        exptime_per_int = (ngroups_per_int)*tframe
        clocktime_per_int = ngroups_per_int*tframe
        eff = (ngroups_per_int - 1.0)/(ngroups_per_int + 1.0)
        nint_per_occultation =  transit_duration/((ngroups_per_int+1.0)*tframe)
        nint_in = np.ceil(nint_per_occultation)
        nint_out = np.ceil(nint_in/expfact_out)
        
    if nint_out < min_nint_trans:
        nint_out = min_nint_trans
   
    timing = {
        "Transit Duration" : (transit_duration)/60.0/60.0,
        "Seconds per Frame" : tframe,
        "Time/Integration incl reset (sec)":clocktime_per_int,
        "APT: Num Groups per Integration" :ngroups_per_int, 
        "Num Integrations Out of Transit":nint_out,
        "Num Integrations In Transit":nint_in,
        "APT: Num Integrations per Occultation":nint_out+nint_in,
        "Observing Efficiency (%)": eff*100.0,
        "Transit+Baseline, no overhead (hrs)": (nint_out+nint_in)*clocktime_per_int/60.0/60.0, 
        "Number of Transits": noccultations
        }      
        
    return timing, {'flag_default':flag_default,'flag_high':flag_high}

def remove_QY(pandeia_dict, instrument):
    """Removes Quantum Yield from Pandeia Fluxes. Place Holder. 
    
    Parameters
    ----------
    pandeia_dict : dict 
        pandeia output dictionary
    instrument : str
        instrument running
    
    Returns
    -------
    dict 
        same exact dictionary with extracted_flux = extracted_flux/QY
    """
    if instrument == 'niriss':
        try:
            qy = fits.open(os.path.join(default_refdata_directory,'jwst', instrument,'qe' ,'jwst_niriss_h2rg_qe_20160902163017.fits'))
        except: 
            raise Exception('PANDEIA REFERENCE DATA NEEDS TO BE UPDATED')

        x_grid = pandeia_dict['1d']['extracted_flux'][0]
        qy_on_grid = np.interp(x_grid, qy[1].data['WAVELENGTH'], qy[1].data['CONVERSION'])
    elif instrument == 'nirspec':
        try:
            qy = fits.open(os.path.join(default_refdata_directory,'jwst', instrument,'qe' ,'jwst_nirspec_qe_20160902193401.fits'))
        except: 
            raise Exception('PANDEIA REFERENCE DATA NEEDS TO BE UPDATED')        
        x_grid = pandeia_dict['1d']['extracted_flux'][0]
        qy_on_grid = np.interp(x_grid, qy[1].data['WAVELENGTH'], qy[1].data['CONVERSION'])
    else:
        #nircam and miri currently have no qy effects
        qy_on_grid = 1.0

    pandeia_dict['1d']['extracted_flux'][1] = pandeia_dict['1d']['extracted_flux'][1]/qy_on_grid
    return pandeia_dict

def perform_out(pandeia_input, pandexo_input,timing, both_spec):
    """Runs pandeia for the out of transit data
    
    Parameters
    ----------
    pandeia_input : dict 
        pandeia specific input info 
    pandexo_input : dict 
        exoplanet specific observation info 
    timing : dict 
        timing dictionary from **compute_timing** 
    both_spec : dict 
        dictionary transit spectra computed from **createInput.bothTrans** 

    Returns
    -------
    dict 
        pandeia output dictionary for out of transit data 
    """
    #pandeia inputs, simulate one integration at a time 
    pandeia_input['configuration']['detector']['ngroup'] = timing['APT: Num Groups per Integration']
    pandeia_input['configuration']['detector']['nint'] = timing['Num Integrations Out of Transit']
    pandeia_input['configuration']['detector']['nexp'] = 1 

    report_out = perform_calculation(pandeia_input, dict_report=False)
    
    return report_out

    
def perform_in(pandeia_input, pandexo_input,timing, both_spec, out, calculation): 
    """Computes in transit data 
    
    Runs Pandeia for the in transit data or computes the in transit simulation 
    from the out of transit pandeia run 
    
    Parameters
    ----------
    pandeia_input : dict 
        pandeia specific input info 
    pandexo_input : dict 
        exoplanet specific observation info 
    timing : dict 
        timing dictionary from **compute_timing** 
    both_spec : dict 
        dictionary transit spectra computed from **createInput.bothTrans** 
    out : dict 
        out of transit dictionary from **perform_in**
    calculation : str
        key which speficies the kind of noise calcualtion 
        (2d extract, slope method, fml, phase_spec). 
        Recommended for transit transmisstion spectra = fml

    Returns
    -------
    dict
        pandeia output dictionary 
    """
    
    #function to run pandeia for in transit
    if calculation == 'phase_spec':
        #return the phase curve since it's all we need 
        report_in = {'time': both_spec['time'],'planet_phase': both_spec['planet_phase']}
    elif calculation == 'fml':
        #for FML method, we only use the flux rate calculated in pandeia so 
        #can compute in transit flux rate without running pandeia a third time     
        report_in = deepcopy(out)
        
        transit_depth = np.interp(report_in['1d']['extracted_flux'][0],
                                    both_spec['wave'], both_spec['frac'])
        report_in['1d']['extracted_flux'][1] = report_in['1d']['extracted_flux'][1]*transit_depth
    else: 
        #only run pandeia a third time if doing slope method and need accurate run for the 
        #nint and timing
        pandeia_input['configuration']['detector']['ngroup'] = timing['APT: Num Groups per Integration']
        pandeia_input['configuration']['detector']['nint'] = timing['Num Integrations In Transit']
        pandeia_input['configuration']['detector']['nexp'] = 1
  
        in_transit_spec = np.array([both_spec['wave'], both_spec['flux_in_trans']])
    
        pandeia_input['scene'][0]['spectrum']['sed']['spectrum'] = in_transit_spec

        report_in = perform_calculation(pandeia_input, dict_report=True)
        instrument = pandeia_input['configuration']['instrument']['instrument']
        #remove QY effects 
        report_in = remove_QY(report_in, instrument)
        report_in.pop('3d')
    
    return report_in
          
def add_warnings(pand_dict, timing, sat_level, flags,instrument): 
    """Add warnings for front end 
    
    Adds in necessary warning flags for a JWST observation usually associated with 
    too few or too many groups or saturation. Alerts user if saturation level is higher 
    than 80 percent and if the number of groups is less than 5. Or, if the full well is 
    greater than 80. These warnings are currently very arbitrary. Will be updated as 
    better JWST recommendations are made. 
    
    Parameters
    ----------
    pand_dict : 
        output from pandeia run 
    timing : dict 
        output from **compute_timing** 
    sat_level : int or float 
        user specified saturation level in fractional (00/100)
    flags : dict 
        warning flags taken from output of **compute_timing**
    instrument : str 
        Only allowable strings are: "nirspec", "niriss", "nircam", "miri"
    
    Returns
    -------
    dict 
        all warnings 

    Notes
    -----    
    These are warnings are just suggestions and are not yet required.   
    """

    ngroups_per_int = timing['APT: Num Groups per Integration']
  
    #check for saturation 
    try:  
        flag_nonl = pand_dict['warnings']['partial_saturated']
    except: 
        flag_nonl = "All good"    
    try: 
        flag_sat = pand_dict['warnings']['full_saturated']
    except: 
        flag_sat = "All good"
        
    #check for too small number of groups
    flag_low = "All good"
    flag_perc = "All good"

    if (sat_level > .80) & (ngroups_per_int <3):
        flag_low = "% full well>80% & only " + str(ngroups_per_int) + " groups"
    if (sat_level > .80): 
        flag_perc = "% full well>80%"

     
    warnings = {
            "Group Number Too Low?" : flag_low,
            "Group Number Too High?": flags["flag_high"],
            "Non linear?" : flag_nonl,
            "Saturated?" : flag_sat,
            "% full well high?": flag_perc, 
            "Num Groups Reset?": flags["flag_default"]
    }

    return warnings     
    
def add_noise_floor(noise_floor, wave_bin, error_spec):
    """Add in noise floor 
    
    This adds in a user speficied noise floor. Does not add the noise floor in quadrature 
    isntead it sets error[error<noise_floor] = noise_floor. If a wavelength dependent 
    noise floor is given and the wavelength ranges are off, it interpolates the out of 
    range noise floor. 
    
    Parameters
    ----------
    noise_floor : str or int 
        file with two column [wavelength, noise(ppm)] or single number with constant noise floor in ppm 
    wave_bin : array of float
        final binned wavelength grid from simulation 
    error_spec : array of float 
        final computed error on the planet spectrum in units of rp^2/r*^2 or fp/f*
    
    Returns
    -------
    array of float 
        error_spec-- new error
    
    Examples
    --------
    
    >>> import numpy as np
    >>> wave = np.linspace(1,2.7,10)
    >>> error = np.zeros(10)+1e-6
    >>> newerror = add_noise_floor(20, wave, error)
    >>> print(newerror)
    [  2.00000000e-05   2.00000000e-05   2.00000000e-05   2.00000000e-05
       2.00000000e-05   2.00000000e-05   2.00000000e-05   2.00000000e-05
       2.00000000e-05   2.00000000e-05]
    """
    #add user specified noise floor 
    if (type(noise_floor)==float) | (type(noise_floor) == int):
        error_spec[error_spec<noise_floor*1e-6] = noise_floor*1e-6
    elif (type(noise_floor)==str):
        read_noise = np.genfromtxt(noise_floor, dtype=(float, float), names='w, n')
        w_overlap = (wave_bin>=min(read_noise['w'])) & (wave_bin<=max(read_noise['w'])) 
        wnoise = wave_bin[w_overlap]
        noise = np.zeros(len(wave_bin))
        noise[w_overlap] = np.interp(wnoise , read_noise['w'], read_noise['n'])
        noise[(wave_bin>max(read_noise['w']))] = read_noise['n'][read_noise['w'] == max(read_noise['w'])]
        noise[(wave_bin<min(read_noise['w']))] = read_noise['n'][read_noise['w'] == min(read_noise['w'])]
        error_spec[error_spec<noise*1e-6] = noise[error_spec<noise*1e-6]*1e-6
    else: 
        raise ValueError('Noise Floor added was not integer or file')
    return error_spec

def bin_wave_to_R(w, R):
    """Creates new wavelength axis at specified resolution
    
    Parameters
    ----------
    w : list of float or numpy array of float
        Wavelength axis to be rebinned 
    R : float or int 
        Resolution to bin axis to 
    
    Returns
    -------
    list of float
        New wavelength axis at specified resolution
    
    Examples
    --------
    
    >>> newwave = bin_wave_to_R(np.linspace(1,2,1000), 10)
    >>> print((len(newwave)))
    11
    """
    wave = []
    tracker = min(w)
    i = 1 
    ind= 0
    firsttime = True
    while(tracker<max(w)):
        if i <len(w)-1:
            dlambda = w[i]-w[ind]
            newR = w[i]/dlambda
            if (newR < R) & (firsttime):
                tracker = w[ind]
                wave += [tracker]
                ind += 1
                i += 1 
                firsttime = True
            elif newR < R:
                tracker = w[ind]+dlambda/2.0
                wave +=[tracker]
                ind = (np.abs(w-tracker)).argmin()
                i = ind+1
                firsttime = True
            else:
                firsttime = False            
                i+=1    
        else:
            tracker = max(w)
            wave += [tracker]
    return np.array(wave)
    
def uniform_tophat_sum(newx,x, y):
    """Adapted from Mike R. Line to rebin spectra
    
    Sums groups of points in certain wave bin 
    
    Parameters
    ----------
    newx : list of float or numpy array of float
        New wavelength grid to rebin to 
    x : list of float or numpy array of float 
        Old wavelength grid to get rid of 
    y : list of float or numpy array of float 
        New rebinned y axis 
    
    Returns
    -------
    array of floats 
        new wavelength grid 
        
    Examples 
    --------
        
    >>> from .pandexo.engine.jwst import uniform_tophat_sum
    >>> oldgrid = np.linspace(1,3,100)
    >>> y = np.zeros(100)+10.0
    >>> newy = uniform_tophat_sum(np.linspace(2,3,3), oldgrid, y)
    >>> newy
    array([ 240.,  250.,  130.])
    """
    newx = np.array(newx)
    szmod=newx.shape[0]
    delta=np.zeros(szmod)
    ynew=np.zeros(szmod)
    delta[0:-1]=newx[1:]-newx[:-1]  
    delta[szmod-1]=delta[szmod-2] 
    #pdb.set_trace()
    for i in range(szmod-1):
        i=i+1
        loc=np.where((x >= newx[i]-0.5*delta[i-1]) & (x < newx[i]+0.5*delta[i]))
        ynew[i]=np.sum(y[loc])
    loc=np.where((x > newx[0]-0.5*delta[0]) & (x < newx[0]+0.5*delta[0]))
    ynew[0]=np.sum(y[loc])
    return ynew

def target_acq(instrument, both_spec, warning): 
    """Contains functionality to compute optimal TA strategy 

    Takes pandexo normalized flux from create_input and checks for saturation, or 
    if SNR is below the minimum requirement for each. Then adds warnings and 2d displays 
    and target acq info to final output dict 

    Parameters
    ----------
    instrument : str 
        possible options are niriss, nirspec, miri and nircam 
    both_spec : dict
        output dictionary from **create_input** 
    warning : dict 
        output dictionary from **add_warnings** 

    Retruns
    -------

    """
    out_spectrum = np.array([both_spec['wave'], both_spec['flux_out_trans']])

    #this automatically builds a default calculation 
    #I got reasonable answers for everything so all you should need to do here is swap out (instrument = 'niriss', 'nirspec','miri' or 'nircam')
    c = build_default_calc(telescope='jwst', instrument=instrument, mode='target_acq', method='taphot')
    c['scene'][0]['spectrum']['sed'] = {'sed_type':'input','spectrum':out_spectrum}
    c['scene'][0]['spectrum']['normalization']['type'] = 'none'
    rphot = perform_calculation(c, dict_report=True)

    #check warnings (pandeia doesn't return values for these warnings, so try will fail if all good)
    try: 
        warnings['TA Satruated?'] = rphot['warnings']['saturated']
    except:
        warnings['TA Satruated?'] = 'All good'

    try: 
        warnings['TA SNR Threshold'] = rphot['warnings']['ta_snr_threshold']
    except:
        warnings['TA SNR Threshold'] = 'All good'

    #build TA dict 
    ta = {'sn':rphot['scalar']['sn'],
            'ngroup': rphot['input']['configuration']['detector']['ngroup'], 
            'saturation':rphot['2d']['saturation']}

def as_dict(out, both_spec ,binned, timing, mag, sat_level, warnings, punit, unbinned,calculation): 
    """Format dictionary for output data 
    
    Takes all output from jwst run and converts it to simple dictionary 
    
    Parameters
    ----------
    out : dict 
        output dictionary from **compute_out**
    both_spec : dict 
        output dictionary from **createInput.bothTrans**
    binned : dict 
        dictionary from **wrapper** 
    timing : dict 
        dictionary from **compute_timing**
    mag : dict 
        magnitude of system 
    sat_level : float or int
        saturation level in electrons 
    warnings : dict 
        warning dictionary from **add_warnings**
    punit : "fp/f*" or "rp^2/r*^2"
        unit of supplied spectra options are: only options are fp/f* or rp^2/r*^2
    unbinned : dict 
        unbinned raw data from **wrapper**
    calculation : str 
        noise calculation type
    
    Returns
    -------
    dict 
        compressed dictionary 
    """
    #for emission spectrum

    p=1.0
    if punit == 'fp/f*': p = -1.0

    timing_div = pd.DataFrame.from_dict(timing, orient='index')
    timing_div.columns = ['Value']
    timing_div = timing_div.to_html()
    timing_div = '<table class="table table-striped"> \n' + timing_div[36:len(timing_div)] 
    timing_div = timing_div.encode()

    warnings_div = pd.DataFrame.from_dict(warnings, orient='index')
    warnings_div.columns = ['Value']
    warnings_div = warnings_div.to_html()
    warnings_div = '<table class="table table-striped"> \n' + warnings_div[36:len(warnings_div)]
    warnings_div = warnings_div.encode()
       
    input_dict = {
   	 "Target Mag": mag , 
   	 "Saturation Level (electons)": sat_level, 
   	 "Instrument": out['input']['configuration']['instrument']['instrument'], 
   	 "Mode": out['input']['configuration']['instrument']['mode'], 
   	 "Aperture": out['input']['configuration']['instrument']['aperture'], 
   	 "Disperser": out['input']['configuration']['instrument']['disperser'], 
   	 "Subarray": out['input']['configuration']['detector']['subarray'], 
   	 "Readmode": out['input']['configuration']['detector']['readmode'], 
 	 "Filter": out['input']['configuration']['instrument']['filter'],
 	 "Primary/Secondary": punit
    }
    
    input_div = pd.DataFrame.from_dict(input_dict, orient='index')
    input_div.columns = ['Value']
    input_div = input_div.to_html()
    input_div = '<table class="table table-striped"> \n' + input_div[36:len(input_div)]
    input_div = input_div.encode()
    
    #add calc type to input dict (doing it here so it doesn't output on webpage
    input_dict["Calculation Type"]= calculation
    
    final_dict = {
    'OriginalInput': {'model_spec':both_spec['model_spec'],
                     'model_wave' : both_spec['model_wave'],
                     'star_spec': both_spec['flux_out_trans']},
    'RawData': unbinned,
    'FinalSpectrum': binned,
        
    #pic output 
    'PandeiaOutTrans': out, 

    #all timing info 
    'timing': timing,
    'warning':warnings,
    'input':input_dict,
    
    #divs for html rendering    
    'timing_div':timing_div, 
    'input_div':input_div,
    'warnings_div':warnings_div,
    }
    return final_dict

    




import numpy as np
from pandeia.engine.instrument_factory import InstrumentFactory
from .pandexo import wrapper
from .load_modes import SetDefaultModes
import os
import pickle as pkl
from joblib import Parallel, delayed
import multiprocessing
import json
from .exomast import get_target_data
from astroquery.simbad import Simbad
import astropy.units as u

user_cores = multiprocessing.cpu_count()

ALL = {"WFC3 G141":False,
       "MIRI LRS":False,
       "NIRISS SOSS":False,
       "NIRSpec G140M":False,
       "NIRSpec G140H":False,
       "NIRSpec G235M":False,
       "NIRSpec G235H":False,
       "NIRSpec G395M":False,
       "NIRSpec G395H":False,
       "NIRSpec Prism":False,
       "NIRCam F322W2":False,
       "NIRCam F444W":False}


def print_instruments(verbose=True):
    """Prints a list of the possible instrument templates to load
    """
    if verbose: print("Choose from the following:")
    return ALL.keys()

def load_exo_dict(planet_name=None):
    """Loads in empty exoplanet dictionary for pandexo input

    Loads in empty exoplanet dictionary so that the user can manually edit different planet
    parameters. Must be done for every bash run. User must edit each keys within the dictionary.
    
    Parameters
    ----------
    planet_name : str
        (Optional) Planet name e.g. 'HD 189733 b' or 'HD189733b'

    Returns
    -------
    dict
        Empty dictionary to be filled out by the user before running PandExo

    Example
    -------
    >>> exo_dict = load_exo_dict()
    >>> exo_dict['planet']['transit_duration'] = 2*60*60 #2 hours
    """
    with open(os.path.join(os.path.dirname(__file__), "reference",
                               "exo_input.json")) as data_file:
        pandexo_input = json.load(data_file)

    if not isinstance(planet_name, type(None)):
        planet_data = get_target_data(planet_name)[0]

        pandexo_input['star']['type'] = 'phoenix'
        pandexo_input['star']['temp'] = planet_data['Teff']
        pandexo_input['star']['metal'] = planet_data['Fe/H']
        pandexo_input['star']['logg'] = planet_data['stellar_gravity']
        Simbad.add_votable_fields('flux(H)')
        Simbad.add_votable_fields('flux(J)')
        star_name = planet_name[:-1]
        jmag = Simbad.query_object(star_name)['FLUX_J'][0]
        if np.ma.is_masked(jmag):
            # Remove 'A' from star_name for systems with binary stars (e.g., WASP-77A)
            star_name = star_name[:-1]
            jmag = Simbad.query_object(star_name)['FLUX_J'][0]
        hmag = Simbad.query_object(star_name)['FLUX_H'][0]

        pandexo_input["star"]["mag"] = jmag
        pandexo_input["star"]["ref_wave"] = 1.25
        pandexo_input["star"]["jmag"] = jmag
        pandexo_input["star"]["hmag"] = hmag
        #optinoal star radius
        pandexo_input["star"]["radius"] = planet_data['Rs']
        pandexo_input["star"]["r_unit"] = planet_data['Rs_unit'][0]+ planet_data['Rs_unit'][1:].lower()

       #optional planet radius/mass
        pandexo_input["planet"]["radius"] = planet_data['Rp']
        pandexo_input["planet"]["r_unit"] = planet_data['Rp_unit'][0]+ planet_data['Rp_unit'][1:].lower()
        pandexo_input["planet"]["mass"] = planet_data['Mp']
        pandexo_input["planet"]["m_unit"] = planet_data['Mp_unit'][0]+ planet_data['Mp_unit'][1:].lower()

        pandexo_input["planet"]["transit_duration"] = planet_data['transit_duration']
        pandexo_input["planet"]["td_unit"] = planet_data['transit_duration_unit']
        depth = pandexo_input["planet"]["radius"]**2 / ((pandexo_input["star"]["radius"]
                                                    *u.Unit(pandexo_input["star"]["r_unit"]) )
                                                        .to(u.Unit(pandexo_input["planet"]["r_unit"]))).value**2
        pandexo_input["planet"]["depth"]      = depth
        if planet_data['inclination'] == None:
            inc = 90
        else:
            inc = planet_data['inclination']

        pandexo_input["planet"]["i"]          = inc
        pandexo_input["planet"]["ars"]        = planet_data['a/Rs']
        period = planet_data['orbital_period']
        period_unit = planet_data['orbital_period_unit']
        pandexo_input["planet"]["period"]     = (period*u.Unit(period_unit)).to(u.Unit('day')).value
        pandexo_input["planet"]["ecc"]        = planet_data['eccentricity']
        pandexo_input["planet"]["ecc"]        = planet_data['eccentricity']
        try:
            pandexo_input["planet"]["w"]      = float(planet_data['omega'] )
        except:
            pandexo_input["planet"]["w"]      = 90.
    return pandexo_input

def load_mode_dict(inst):
    """Function to pull in correct instrument dictionary

    This is the instrument counterpart to load_exo_dict. It loads in a template instrument
    dictionary for a specific instrument mode.

    Parameters
    ----------
    inst : str
        One of the allowable instrument keys. To see which instrument modes are available
        use `print_instruments()`

    Returns
    -------
    dict
        Filled out template of instrument dictionary, which can be editted before
        running to PandExo (not required).

    Example
    -------
    >>> inst_dict = load_mode_dict('MIRI LRS')
    >>> inst_dict['configuration']['instrument']['aperture'] = 'lrsslit'
    """
    return SetDefaultModes(inst).pick()

def get_thruput(inst, niriss=1, nirspec='f100lp'):
    """Returns complete instrument photon to electron conversion efficiency
    Pulls complete instrument photon to electron conversion efficiency
    (PCE) based on instrument key input

    Parameters
    ----------
    inst : str
        One of the instrument keys in `print_instruments`
    niriss : int
        (Optional) defines which niriss order you want (1 or 2)
    nirspec : str
        (Optional) for NIRISS G140M/H there are two available filters (f100lp and f070lp)
        if you are selecting G140M or G140H, this allows you to pick which one
    Returns
    -------
    dict
        Dictionary with wave solution and PCE

    Example
    -------
    >>> thru_dict = get_thruput('NIRISS SOSS_Or1')
    """

    #pull correct dictionary
    input_dict =  SetDefaultModes(inst).pick()
    conf = input_dict['configuration']
    conf['detector']['ngroup'] = 2

    if conf['instrument']['instrument'].lower() =='niriss':
        #pandeia handles slit losses inside the 2d engine. So, you need to account for the
        #extra .663 here
        conf["instrument"]["disperser"] = conf["instrument"]["disperser"] +'_'+str(niriss)
        i = InstrumentFactory(config=conf)
        wr = i.get_wave_range()
        wave = np.linspace(wr['wmin'], wr['wmax'], num=500)
        pce = i.get_total_eff(wave)
        return {'wave':wave,'pce':0.663*pce}
    elif (conf['instrument']['instrument'].lower() =='nirspec') and ('g140' in conf["instrument"]["disperser"]):
        conf["instrument"]["filter"] = nirspec

    i = InstrumentFactory(config=conf)
    wr = i.get_wave_range()
    wave = np.linspace(wr['wmin'], wr['wmax'], num=500)
    pce = i.get_total_eff(wave)
    return {'wave':wave,'pce':pce}

def run_param_space(i,exo,inst,param_space, verbose=False):
    """Changes exo dictionary and submits run

    This function is used to reset the exo dictionary based on what parameter
    is being looped over and submits run to `wrapper` so that all the jobs
    are run in parallel

    Parameters
    ----------
    i : str or float
        Can be either a str or float based on what you are looping through (str for
        filenames, float for stellar temps, float for magnitudes, etc)
    exo : dict
        Exoplanet dictionary which can be loaded in and editted through `load_exo_dict`
    inst : str
        Key which indicates with instrument
    param_space : str
        Set of keys within exo_dict to indicate which parameter to loop through.
        Should be in the format of "first level of dict"+"second level of dict".
        For example, for stellar temp `param_space` would be "star+temp"
    verbose : bool 
        (Optional) prints out checkpoints. Assuming the user does not want a load of 
        print statements for parallel runs

    Returns
    -------
    dict
        Dictionary with output of pandexo. Key is the value of the parameter that was
        looped through.
    """
    #break up parameter space to two separate dictionary keys
    key1 = param_space[0:param_space.find('+')]
    key2 = param_space[param_space.find('+')+1:len(param_space)]
    exo[key1][key2] = i
    #load in correct dict format
    inst_dict = load_mode_dict(inst)
    name = os.path.split(str(i))[1]
    return {name: wrapper({"pandeia_input": inst_dict , "pandexo_input":exo}, verbose=verbose)}

def run_inst_space(inst,exo, verbose=False):
    """Changes inst dictionary and submits run

    This function is used to reset the instrument dictionary.

    Parameters
    ----------
    exo : dict
        Exoplanet dictionary which can be loaded in and editted through `load_exo_dict`
    inst : str
        Key which indicates with instrument
    verbose : bool 
        (Optional) prints out checkpoints. Assuming the user does not want a load of 
        print statements for parallel runs

    Returns
    -------
    dict
        Dictionary with output of pandexo. Key is the value of the parameter that was
        looped through.
    """
    #load in correct dict format
    inst_dict = load_mode_dict(inst)
    return {inst: wrapper({"pandeia_input": inst_dict , "pandexo_input":exo}, verbose=verbose)}


def run_pandexo(exo, inst, param_space = 0, param_range = 0,save_file = True,
                            output_path=os.getcwd(), output_file = '',num_cores=user_cores, verbose=True):
    """Submits multiple runs of pandexo in parallel.

    Functionality: program contains functionality for running single or
    multiple runs of PandExo

    Parameters
    ----------
    exo : dict
        exoplanet input dictionary
    inst : dict or str or list of str
        instrument input dictionary OR LIST of keys (for allowable keys see `print_instruments()`
    param_space : str or 0
        (Optional) Default is 0 = no exoplanet parameter space. To run through a parameter
        specify which one need to specify two keys from exo dict with + in between.
        i.e. observation+fraction
        star+temp
        planet+exopath
    param_range : list of str or list of float
        (Optional) Default = 0 An array or list over which to run the parameters space.
        i.e. array of temperatures if running through stellar temp or
        array of files if running through planet models. Must specify param_space
        if using this.
    save_file : bool
        (Optional) Default = True saves file, False does not
    output_path : str
        (Optional) Defaults to current working directory
    output_file : str
        (Optional) Default is "singlerun.p" for single runs, "param_space.p" for exo parameter runs
        or "instrument_run.p" for instrument parameter space runs.
    verbose : bool 
        (Optional) For single runs, if false, it turns off all print statements. For parameter space 
        runs it is defaulted to never print statements out.

    Returns
    -------
    dict
        For single run output will just be a single PandExo output dictionary
        https://github.com/natashabatalha/PandExo/wiki/PandExo-Output
        For multiple runs the output will be organized into a list with each
        a dictionary named by whatever you are looping through
        i.e. [{'First temp': PandExoDict}, {'Second temp': PandExoDict}, etc..]

    Example
    -------
    For single run:

    >>> a = run_pandexo(exo_dict, ['MIRI LRS'])

    For multiple instruments:

    >>> a = run_pandexo(exo_dict, ['MIRI LRS','NIRSpec G395H'])

    Loop through a exoplanet parameter (stellar magnitude):

    >>> a = run_pandexo(exo_dict, ['NIRSpec G395M'],
            param_space ='star+mag',param_range = np.linspace(6,10,5))
    """

    #single instrument mode with dictionary input OR single planet
    if type(inst) == dict:
        if verbose: print("Running Single Case w/ User Instrument Dict")
        results =wrapper({"pandeia_input": inst , "pandexo_input":exo}, verbose=verbose)
        if output_file == '':
            output_file = 'singlerun.p'
        if save_file: pkl.dump(results, open(os.path.join(output_path,output_file),'wb'))
        return results

    #make sure inst is in list format.. makes my life so much easier
    try:
        if type(inst) != list:
            raise ValueError
    except ValueError:
        print('Instrument input is not dict so must be list')
        print('Enter in format ["NIRSpec G140M"] or ["NIRISS SOSS","MIRI LRS"]')
        return

    #single instrument mode and single planet OR several planets

    if len(inst)==1 and inst[0] != 'RUN ALL':

        #start case of no parameter space run
        if isinstance(param_space, (float, int)) or isinstance(param_range, (float, int)):
            if verbose: print("Running Single Case for: " + inst[0])
            inst_dict = load_mode_dict(inst[0])
            results =wrapper({"pandeia_input": inst_dict , "pandexo_input":exo}, verbose=verbose)
            if output_file == '':
                output_file = 'singlerun.p'
            if save_file: pkl.dump(results, open(os.path.join(output_path,output_file),'wb'))
            return results

        #if there are parameters to cycle through this will run
        if verbose: print("Running through exo parameters in parallel: " + param_space)
        #run the above function in parallel
        results = Parallel(n_jobs=num_cores)(delayed(run_param_space)(i,exo,inst[0],param_space) for i in param_range)

        #Default dump all results [an array of dictionaries] into single file
        #and return results immediately to user
        if output_file == '':
            output_file = param_space + '.p'

        if save_file: pkl.dump(results, open(os.path.join(output_path,output_file),'wb'))
        return results

    #run several different instrument modes and single planet
    if verbose: print("Running select instruments")
    if len(inst)>1:

        results = Parallel(n_jobs=num_cores)(delayed(run_inst_space)(i, exo) for i in inst)

        #Default dump all results [an array of dictionaries] into single file
        #and return results immediately to user
        if output_file == '':
            output_file =  'instrument_run.p'
        if save_file: pkl.dump(results, open(os.path.join(output_path,output_file),'wb'))
        return results

    #cycle through all options
    elif inst[0].lower() == 'run all':
        if verbose: print("Running through all instruments")
        results = Parallel(n_jobs=num_cores)(delayed(run_inst_space)(i, exo) for i in ALL.keys())

        #Default dump all results [an array of dictionaries] into single file
        #and return results immediately to user
        if output_file == '':
            output_file =  'instrument_run.p'
        if save_file: pkl.dump(results, open(os.path.join(output_path,output_file),'wb'))
        return results

def subarrays(inst):
  """function to show availalble subarrays and their times (in secons)

  Parameters
  ----------
  inst : str
    string of either niriss, nirspec, miri or nircam

  Returns
  -------
  dict
    dictionary with name of subarray as keys and time in seconds as entry
  """
  print("Subarray field stored in inst_dict['configuration']['detector']['subarray']")

  if inst.lower() == 'niriss':
    return {'substrip96':2.2129,'substrip256':5.4913}
  elif inst.lower() == 'nirspec':
    return {'sub1024a':0.451,'sub1024b':0.451,'sub2048':0.90156,'sub512':0.22572}
  elif inst.lower() == 'miri':
    return {'slitlessprism':0.159}
  elif inst.lower()  == 'nircam':
    return {"subgrism64":0.34, "subgrism128":0.67, "subgrism256":1.34,
      "subgrism64 (noutputs=1)":1.3, "subgrism128 (noutputs=1)":2.6, "subgrism256 (noutputs=1)":5.2}
  else:
    raise Exception("Only instruments are niriss, nirspec, miri, nircam. Pick one.")

def dispersers(inst):
  """function to show available dispersers

  Parameters
  ----------
  inst : str
    string of either niriss, nirspec, miri or nircam

  Returns
  -------
  list
    lsit with available dispersers
  """
  print("Dispersers field stored in inst_dict['configuration']['instrument']['disperser']")
  if inst.lower() == 'niriss':
    return ['gr700xd']
  elif inst.lower() == 'nirspec':
    return ['g140m','g140h','g235m','g235h','g395m','g395h','prism']
  elif inst.lower() == 'miri':
    return ['p750l']
  elif inst.lower()  == 'nircam':
    return ['grismr']
  else:
    raise Exception("Only instruments are niriss, nirspec, miri, nircam. Pick one.")

def filters(inst):
    """Function to show availalbe filters

  Parameters
  ----------
  inst : str
    string of either niriss, nirspec, miri or nircam

  Returns
  -------
  list
    list with availalbe filters
  """
    print("Filters field stored in inst_dict['configuration']['instrument']['filter']")

    if inst.lower() == 'niriss':
        print("No filters for niriss soss, type None, or null in filter field")
        return [None]
    elif inst.lower() == 'nirspec':
        return ['f070lp','f100lp','f170lp','f290lp','clear']
    elif inst.lower() == 'miri':
        print("No filters for miri lrs, type None, or null in filter field")
        return [None]
    elif inst.lower()  == 'nircam':
        return ['f322w2','f444w']
    else:
        raise Exception("Only instruments are niriss, nirspec, miri, nircam. Pick one.")

def grid_options(grid = 'fortney'):
    """Function to show available grid options

    PandExo now supports various grid options. Currently, the only one that is availalbe
    is the Fortney grid for giant planets. We will be implementing others, as they
    become available. It will become increasingly difficult for users to see what
    options are availalbe to them. This function should guide users to choose
    the grid that best fits their needs.

    Parameters
    ----------
    grid : str
        (Optional) string which 'fortney'

    """
    return



import os
import json
import numpy as np


class SetDefaultModes():
    """Class to load default instrument mode dicts
    
    This class contains functionality for loading observing modes for exoplanet observations 
    This is NOT a complete set of ALL possibleobserving modes. Instead, it offers a starting point for 
    choosing one instrument specification. There is one function for each instrument. For example, 
    It sets slitless mode for MIRI LRS. If users are interested in other specific observation modes 
    they should load in the dictionary and then edit individual keys. 
    
    Included modes are: 
       - "MIRI LRS"
       - "NIRISS SOSS"
       - "NIRSpec G140M"
       - "NIRSpec G140H"
       - "NIRSpec G235M"
       - "NIRSpec G235H"
       - "NIRSpec G395M"
       - "NIRSpec G395H"
       - "NIRSpec Prism"
       - "NIRCam F322W2"
       - "NIRCam F444W"
       - "WFC3 G102"
       - "WFC3 G141"
       
    Parameters
    ----------
    inst : str
        Allowable strings listed above 
    """

    def __init__(self, inst):
        self.instrument = inst[0:inst.find(' ')].lower()
        self.config = inst[inst.find(' ')+1:len(inst)].lower()   
    
    def pick(self): 
        """Points to specific instrument based on key choice
        """
        try: 
            return getattr(self, self.instrument)()
        except: 
            print("INVALID INSTRUMENT NAME")
            return 
                                   
    def wfc3(self):
        """Handles WFC3 template
        """
        #wfc3_input
        with open(os.path.join(os.path.dirname(__file__), "reference",
                               "wfc3_input.json")) as data_file:
            pandeia_data = json.load(data_file)
            pandeia_data["configuration"]["instrument"]["disperser"] = self.config
        return pandeia_data
    
    def niriss(self):
        """Handles NIRISS template
        """
        #need to add in functionality for two orders, right now you only have option to do sub80
        
        with open(os.path.join(os.path.dirname(__file__), "reference",
                               "niriss_input.json")) as data_file:
            pandeia_data = json.load(data_file)
        return pandeia_data

    def nirspec(self):
        """Handles NIRSpec template
        """
        filters = {'g140m':'f100lp','g140h':'f100lp',
                    'g235m':'f170lp','g235h':'f170lp',
                    'g395m':'f290lp','g395h':'f290lp',
                    'prism': 'clear'
                    }
        with open(os.path.join(os.path.dirname(__file__), "reference",
                               "nirspec_input.json")) as data_file:
            pandeia_data = json.load(data_file)
            pandeia_data["configuration"]["instrument"]["disperser"] = self.config
            pandeia_data["configuration"]["instrument"]["filter"] = filters[self.config]    
            if self.config == 'prism':
                pandeia_data["configuration"]["detector"]["subarray"] = 'sub512'

        return pandeia_data

    def nircam(self):
        """Handles NIRCam template
        """
        with open(os.path.join(os.path.dirname(__file__), "reference",
                               "nircam_input.json")) as data_file:
            pandeia_data = json.load(data_file)
            pandeia_data["configuration"]["instrument"]["filter"]  = self.config           
        return pandeia_data

    def miri(self):
        """Handles MIRI template
        """
        #only lrs slit less     
        with open(os.path.join(os.path.dirname(__file__), "reference",
                               "miri_input.json")) as data_file:
                pandeia_data = json.load(data_file)
        return pandeia_data



import numpy as np
import pandas as pd
from .create_input import hst_spec
import batman
from .RECTE import RECTE
from scipy import optimize

def wfc3_GuessNOrbits(trdur):
    '''Predict number of HST orbits

    Predict number of HST orbits for transit observation when not provided by the user.

    Parameters
    ----------
    trdur : float
        transit duration in days

    Returns
    -------
    float
        number of requested orbits per transit (including discarded thermal-settling orbit)
    '''
    # Compute # of HST orbits during transit
    # ~96 minutes per HST orbit
    orbitsTr = trdur*24.*60/96
    if orbitsTr <= 1.5:
        norbits = 4.
    elif orbitsTr <= 2.0:
        norbits = 5.
    else:
        norbits = np.ceil(orbitsTr*2+1)

    return norbits


def wfc3_GuessParams(jmag, disperser, scanDirection, subarray, obsTime, maxScanHeight=180., maxExptime=150., targetFluence=30000., hmag=None):
    '''Predict nsamp and samp_seq when values not provided by the user.

    Parameters
    ----------
    jmag : float
        J-band magnitude
    disperser : str
        grism ('G141' or 'G102')
    scanDirection : str
        spatial scan direction ('Forward' or 'Round Trip')
    subarray : str
        Subarray aperture ('grism256' or 'grism512')
    obsTime : float
        Available observing time per HST orbit in seconds
    maxScanHeight : float
        (optional) maximum scan height in pixels
    maxExptime : float
        (Optional) default=150.0, maximum exposure time in seconds
    targetFluence : float
        (Optional) Desired fluence in electrons per pixel
    hmag : float
        (Optional) H-band magnitude

    Returns
    -------
    float
        nsamp--number of up-the-ramp samples (1..15)
    str
        samp_seq--time between non-destructive reads
    '''
    allnsamp = np.arange(1, 16)
    allsampseq = ['spars5', 'spars10', 'spars25']
    maxDutyCycle = 0
    for samp_seq in allsampseq:
        for nsamp in allnsamp:
            exptime, tottime, scanRate, scanHeight, fluence = wfc3_obs(jmag, disperser, scanDirection,
                                                                       subarray, nsamp, samp_seq, targetFluence, hmag)
            # Compute duty cycle and compare
            # Exposure time should be less than 2.5 minutes to achieve good time resolution
            ptsOrbit = np.floor(obsTime/tottime)
            dutyCycle = (exptime*(ptsOrbit-1))/50./60*100
            if (dutyCycle > maxDutyCycle) and (exptime < maxExptime) and (scanHeight < maxScanHeight):
                maxDutyCycle = dutyCycle
                bestsampseq = samp_seq
                bestnsamp = nsamp

    return bestnsamp, bestsampseq


def wfc3_obs(jmag, disperser, scanDirection, subarray, nsamp, samp_seq, targetFluence=30000., hmag=None):
    '''Determine the recommended exposure time, scan rate, scan height, and overheads.

    Parameters
    ----------
    jmag : float
        J-band magnitude
    disperser : str
        Grism ('G141' or 'G102')
    scanDirection : str
        spatial scan direction ('Forward' or 'Round Trip')
    subarray : str
        Subarray aperture ('grism256' or 'grism512')
    nsamp : float
        Number of up-the-ramp samples (1..15)
    samp_seq : str
        Time between non-destructive reads ('SPARS5', 'SPARS10', or 'SPARS25')
    targetFluence : float
        (Optional) Desired fluence in electrons per pixel
    hmag : float
        (Optional) H-band magnitude

    Returns
    -------
    float
        exptime--exposure time in seconds
    float
        tottime--total frame time including overheads in seconds
    float
        scanRate--recommented scan rate in arcsec/s
    float
        scanHeight--scan height in pixels
    float
        fluence--maximum pixel fluence in electrons
    '''
    # Estimate exposure time
    if subarray == 'grism512':
        # GRISM512
        if samp_seq == 'spars5':
            exptime = 0.853 + (nsamp-1)*2.9215  # SPARS5
        elif samp_seq == 'spars10':
            exptime = 0.853 + (nsamp-1)*7.9217  # SPARS10
        elif samp_seq == 'spars25':
            exptime = 0.853 + (nsamp-1)*22.9213  # SPARS25
        else:
            print(("****HALTED: Unknown SAMP_SEQ: %s" % samp_seq))
            return
    else:
        # GRISM256
        if samp_seq == 'spars5':
            exptime = 0.280 + (nsamp-1)*2.349  # SPARS5
        elif samp_seq == 'spars10':
            exptime = 0.278 + (nsamp-1)*7.3465  # SPARS10
        elif samp_seq == 'spars25':
            exptime = 0.278 + (nsamp-1)*22.346  # SPARS25
        else:
            print(("****HALTED: Unknown SAMP_SEQ: %s" % samp_seq))
            return

    # Recommended scan rate
    if hmag == None:
        hmag = jmag
    #scanRate = np.round(1.9*10**(-0.4*(hmag-5.9)), 3)  # arcsec/s
    #scanRate = np.round(2363./targetFluence*10**(-0.4*(jmag-9.75)), 3) # arcsec/s
    scanRate = (2491./targetFluence)*10**(-0.4*(jmag-9.75)) - (161/targetFluence)*10**(-0.4*(jmag-hmag))
    if disperser == 'g102':
        # G102/G141 flux ratio is ~0.8
        scanRate *= 0.8
    # Max fluence in electrons/pixel
    #fluence = (5.5/scanRate)*10**(-0.4*(hmag-15))*2.4  # electrons
    #fluence = (2363./scanRate)*10**(-0.4*(jmag-9.75))  # electrons
    fluence = (2491./scanRate)*10**(-0.4*(jmag-9.75)) - (161/scanRate)*10**(-0.4*(jmag-hmag))
    if disperser == 'g102':
        # WFC3_ISR_2012-08 states that the G102/G141 scale factor is 0.96 DN/electron
        fluence *= 0.96
        # G102/G141 flux ratio is ~0.8
        fluence *= 0.8
    # Scan height in pixels
    scanRatep = scanRate/0.121  # pixels/s
    scanHeight = scanRatep*exptime  # pixels
    '''
    #Quadratic correlation between scanRate and read overhead
    foo     = np.array([0.0,0.1,0.3,0.5,1.0,2.0,3.0,4.0])/0.121
    bar     = np.array([ 40, 40, 41, 42, 45, 50, 56, 61])
    c       = np.polyfit(foo,bar,2)
    model   = c[2] + c[1]*foo + c[0]*foo**2
    #c = [  6.12243227e-04,   6.31621064e-01,   3.96040946e+01]
    '''
    # Define instrument overheads (in seconds)
    c = [6.12243227e-04, 6.31621064e-01, 3.96040946e+01]
    read = c[2] + c[1]*scanRatep + c[0]*scanRatep**2
    # Correlation between scanHeight/scanRate and pointing overhead was determined elsewhere
    if scanDirection == 'Round Trip':
        # Round Trip scan direction doesn't have to return to starting point, therefore no overhead
        pointing = 0.
    elif scanDirection == 'Forward':
        c = [3.18485340e+01,   3.32968829e-02,   1.65687590e-02,
             7.65510038e-01,  -6.24504499e+01,   5.51452028e-03]
        pointing = c[0]*(1 - np.exp(-c[2]*(scanHeight-c[4]))) + \
            c[1]*scanHeight + c[3]*scanRatep + c[5]*scanRatep**2
    else:
        print(("****HALTED: Unknown scan direction: %s" % scanDirection))
        return
    # Estimate total frame time including overheads
    tottime = exptime+read+pointing  # seconds

    return exptime, tottime, scanRate, scanHeight, fluence


def wfc3_TExoNS(dictinput):
    '''Compute Transit depth uncertainty

    Compute the transit depth uncertainty for a defined system and number of spectrophotometric channels.
    Written by Kevin Stevenson      October 2016

    Parameters
    ----------
    dictinput : dict
        dictionary containing instrument parameters and exoplanet specific parameters. {"pandeia_input":dict1, "pandexo_input":dict1}

    Returns
    -------
    float
        deptherr--transit depth uncertainty per spectrophotometric channel
    float
        chanrms--light curve root mean squarerms
    float
        ptsOrbit--number of HST frames per orbit
    '''
    pandeia_input = dictinput['pandeia_input']
    pandexo_input = dictinput['pandexo_input']

    jmag = pandexo_input['star']['jmag']
    if np.ma.is_masked(jmag):
        print("Jmag not found.")
    try:
        hmag = pandexo_input['star']['hmag']
    except:
        hmag = jmag
        print("Hmag not found. Assuming no color dependence in the stellar type.")
    trdur = pandexo_input['planet']['transit_duration']
    numTr = pandexo_input['observation']['noccultations']

    schedulability = pandeia_input['strategy']['schedulability']
    scanDirection = pandeia_input['strategy']['scanDirection']
    nchan = pandeia_input['strategy']['nchan']
    norbits = pandeia_input['strategy']['norbits']
    useFirstOrbit = pandeia_input['strategy']['useFirstOrbit']
    try:
        targetFluence = pandeia_input['strategy']['targetFluence']
    except:
        targetFluence = 30000.
        print("Assuming a target fluence of 30,000 electrons.")
    disperser = pandeia_input['configuration']['instrument']['disperser'].lower(
    )
    subarray = pandeia_input['configuration']['detector']['subarray'].lower()
    nsamp = pandeia_input['configuration']['detector']['nsamp']
    samp_seq = pandeia_input['configuration']['detector']['samp_seq']

    try:
        samp_seq = samp_seq.lower()
    except:
        pass

    if disperser == 'g141':
        # Define reference Jmag, flux, variance, and exposure time for GJ1214
        refmag = 9.750
        refflux = 2.32e8
        refvar = 2.99e8
        refexptime = 88.436
    elif disperser == 'g102':
        # Define reference Jmag, flux, variance, and exposure time for WASP12
        refmag = 10.477
        refflux = 8.26e7
        refvar = 9.75e7
        refexptime = 103.129
    else:
        print(("****HALTED: Unknown disperser: %s" % disperser))
        return

    # Determine max recommended scan height
    if subarray == 'grism512':
        maxScanHeight = 430
    elif subarray == 'grism256':
        maxScanHeight = 180
    else:
        print(("****HALTED: Unknown subarray aperture: %s" % subarray))
        return

    # Define maximum frame time
    maxExptime = 150.

    # Define available observing time per HST orbit in seconds
    if str(schedulability) == '30':
        obsTime = 51.3*60
    elif str(schedulability) == '100':
        obsTime = 46.3*60
    else:
        print(("****HALTED: Unknown schedulability: %s" % schedulability))
        return

    # Compute recommended number of HST orbits and compare to user specified value
    guessorbits = wfc3_GuessNOrbits(trdur)
    if norbits == None:
        norbits = guessorbits
    elif norbits != guessorbits:
        print(("****WARNING: Number of specified HST orbits does not match number of recommended orbits: %0.0f" % guessorbits))

    if nsamp == 0 or nsamp == None or samp_seq == None or samp_seq == "none":
        # Estimate reasonable values
        nsamp, samp_seq = wfc3_GuessParams(jmag, disperser, scanDirection, subarray, obsTime, maxScanHeight, maxExptime, targetFluence, hmag)

    # Calculate observation parameters
    exptime, tottime, scanRate, scanHeight, fluence = wfc3_obs(jmag, disperser, scanDirection, subarray,
                                                               nsamp, samp_seq, targetFluence, hmag=hmag)
    if scanHeight > maxScanHeight:
        print(("****WARNING: Computed scan height exceeds maximum recommended height of %0.0f pixels." % maxScanHeight))
    if exptime > maxExptime:
        print(("****WARNING: Computed frame time (%0.0f seconds) exceeds maximum recommended duration of %0.0f seconds." % (exptime, maxExptime)))

    # Compute number of data points (frames) per orbit
    ptsOrbit = np.floor(obsTime/tottime)
    # First point (frame) is always low, ignore when computing duty cycle
    dutyCycle = (exptime*(ptsOrbit-1))/50./60*100

    # Compute number of non-destructive reads per orbit
    readsOrbit = ptsOrbit*(nsamp+1)

    # Look for mid-orbit buffer dumps
    if (subarray == 'grism256') and (readsOrbit >= 300) and (exptime <= 43):
        print(
            "****WARNING: Observing plan may incur mid-orbit buffer dumps.  Check with APT.")
    if (subarray == 'grism512') and (readsOrbit >= 120) and (exptime <= 100):
        print(
            "****WARNING: Observing plan may incur mid-orbit buffer dumps.  Check with APT.")

    # Compute number of HST orbits per transit
    # ~96 minutes per HST orbit
    orbitsTr = trdur*24.*60/96

    # Estimate number of good points during planet transit
    # First point in each HST orbit is flagged as bad; therefore, subtract from total

    if orbitsTr < 0.5:
        # Entire transit fits within one HST orbit
        ptsInTr = ptsOrbit * orbitsTr/0.5 - 1
    elif orbitsTr <= 1.5:
        # Assume one orbit centered on mid-transit time
        ptsInTr = ptsOrbit - 1
    elif orbitsTr < 2.:
        # Assume one orbit during full transit and one orbit during ingress/egress
        ptsInTr = ptsOrbit * \
            (np.floor(orbitsTr) +
             np.min((1, np.remainder(orbitsTr-np.floor(orbitsTr)-0.5, 1)/0.5))) - 2
    else:
        # Assume transit contains 2+ orbits timed to maximize # of data points.
        ptsInTr = ptsOrbit * (np.floor(orbitsTr) + np.min(
            (1, np.remainder(orbitsTr-np.floor(orbitsTr), 1)/0.5))) - np.ceil(orbitsTr)

    # Estimate number of good points outside of transit
    # Discard first HST orbit
    ptsOutTr = (ptsOrbit-1) * (norbits-1) - ptsInTr

    # Compute transit depth uncertainty per spectrophotometric channel
    ratio = 10**((refmag - jmag)/2.5)
    flux = ratio*refflux*exptime/refexptime
    fluxvar = ratio*refvar*exptime/refexptime
    chanflux = flux/nchan
    chanvar = fluxvar/nchan
    chanrms = np.sqrt(chanvar)/chanflux*1e6  # ppm
    inTrrms = chanrms/np.sqrt(ptsInTr*numTr)  # ppm
    outTrrms = chanrms/np.sqrt(ptsOutTr*numTr)  # ppm
    deptherr = np.sqrt(inTrrms**2 + outTrrms**2)  # ppm

    info = {"Number of HST orbits": norbits,
            "Use first orbit":   useFirstOrbit,
            "WFC3 parameters: NSAMP": nsamp,
            "WFC3 parameters: SAMP_SEQ": samp_seq.upper(),
            "Scan Direction": scanDirection,
            "Recommended scan rate (arcsec/s)": scanRate,
            "Scan height (pixels)": scanHeight,
            "Maximum pixel fluence (electrons)": fluence,
            "exposure time": exptime,
            "Estimated duty cycle (outside of Earth occultation)": dutyCycle,
            "Transit depth uncertainty(ppm)": deptherr,
            "Number of channels": nchan,
            "Number of Transits": numTr}

    return {"spec_error": deptherr/1e6,
            "light_curve_rms": chanrms/1e6,
            "nframes_per_orb": ptsOrbit,
            "info": info}


def calc_start_window(eventType, rms, ptsOrbit, numOrbits, depth, inc, aRs, period, windowSize, ecc=0, w=90., duration=None, offset=0., useFirstOrbit=False):
    '''Calculate earliest and latest start times

    Plot earliest and latest possible spectroscopic light curves for given start window size

    Parameters
    ----------
    eventType : str
        'transit' or 'eclipse'
    rms : float
        light curve root-mean-square
    ptsOrbit : int
        number of frames per HST orbit
    numOrbits :  int
        number of HST orbits per visit
    depth : float
        transit/eclipse depth
    inc : float
        orbital inclination in degrees
    aRs : float
        Semi-major axis in units of stellar radii (a/R*)
    period : float
        orbital period in days
    windowSize : float
        observation start window size in minutes
    ecc : float
        (Optional) eccentricity
    w : float
        (Optional) longitude of periastron in degrees
    duration : float
        (Optional) full transit/eclipse duration in days
    offset : float
        (Optional) manual offset in observation start time, in minutes
    useFirstOrbit : bool
        (Optional) whether to use first orbit

    Returns
    -------
    float
        minphase--earliest observation start phase
    float
        maxphase--latest observation start phase
    '''
    ptsOrbit  = int(ptsOrbit)
    numOrbits = int(numOrbits)

    hstperiod = 96./60/24                 # HST orbital period, days
    punc = windowSize/120./24/period  # Half start window size, in phase
    cosi = np.cos(inc*np.pi/180)     # Cosine of the inclination
    rprs = np.sqrt(depth)            # Planet-star radius ratio

    params = batman.TransitParams()
    if eventType == 'transit':
        midpt = period
        b = aRs*cosi*(1-ecc**2)/(1+ecc*np.sin(w*np.pi/180))  # Impact parameter
        # Account for planet speed on eccentric orbits
        sfactor = np.sqrt(1-ecc**2)/(1+ecc*np.sin(w*np.pi/180))
        # limb darkening coefficients
        params.u = [0.1, 0.1]
    elif eventType == 'eclipse':
        #midpt = period/2*(1+4*ecc*np.cos(w*np.pi/180)/np.pi)
        midpt = calculate_tsec(period, ecc, w*np.pi/180, inc*np.pi/180, t0=0, tperi=None, winn_approximation=False)
        b = aRs*cosi*(1-ecc**2)/(1-ecc*np.sin(w*np.pi/180))  # Impact parameter
        # Account for planet speed on eccentric orbits
        sfactor = np.sqrt(1-ecc**2)/(1-ecc*np.sin(w*np.pi/180))
        # limb darkening coefficients
        params.u = [0.0, 0.0]
    else:
        print(("****HALTED: Unknown event type: %s" % eventType))
        return
    params.t0 = midpt/period          # phase of transit/eclipse
    params.per = 1.                    # orbital period, units are orbital phase
    # planet radius (in units of stellar radii)
    params.rp = rprs
    # semi-major axis (in units of stellar radii)
    params.a = aRs
    params.inc = inc                   # orbital inclination (in degrees)
    params.ecc = ecc                   # eccentricity
    params.w = w                     # longitude of periastron (in degrees)
    params.limb_dark = "quadratic"           # limb darkening model

    # Transit/eclipse duration (in days)
    if duration == None:
        duration = period/np.pi * \
            np.arcsin(
                1./aRs*np.sqrt(((1+rprs)**2-(aRs*cosi)**2)/(1-cosi**2)))*sfactor
    phase1 = (midpt + duration/2. - hstperiod*(numOrbits-2 -
                                               useFirstOrbit) - hstperiod/2 + offset/24./60)/period
    phase2 = (midpt - duration/2. - hstperiod*2 + offset/24./60)/period
    minphase = (phase1+phase2)/2-punc
    maxphase = (phase1+phase2)/2+punc

    # Compute light curves at extremes of HST start window
    npts = int(4 * ptsOrbit * numOrbits)
    phdur = duration/period
    
    phase1 = np.linspace(minphase+(1-useFirstOrbit)*hstperiod/period,
                         minphase+hstperiod/period*(numOrbits-1)+hstperiod/period/2, int(npts))
    phase2 = np.linspace(maxphase+(1-useFirstOrbit)*hstperiod/period,
                         maxphase+hstperiod/period*(numOrbits-1)+hstperiod/period/2, int(npts))
    m = batman.TransitModel(params, phase1)
    trmodel1 = m.light_curve(params)
    m = batman.TransitModel(params, phase2)
    trmodel2 = m.light_curve(params)
    obsphase1 = []
    obsphase2 = []
    for i in range(numOrbits):
        obsphase1 = np.r_[obsphase1, np.linspace(
            minphase+hstperiod/period*i, minphase+hstperiod/period*i+hstperiod/period/2, int(ptsOrbit))]
        obsphase2 = np.r_[obsphase2, np.linspace(
            maxphase+hstperiod/period*i, maxphase+hstperiod/period*i+hstperiod/period/2, int(ptsOrbit))]
    m = batman.TransitModel(params, obsphase1)
    obstr1 = m.light_curve(params) + np.random.normal(0, rms, obsphase1.shape)
    m = batman.TransitModel(params, obsphase2)
    obstr2 = m.light_curve(params) + np.random.normal(0, rms, obsphase2.shape)

    return {'obsphase1': obsphase1, 'light_curve_rms': rms, 'obstr1': obstr1, 'obsphase2': obsphase2,
            'obstr2': obstr2, 'minphase': minphase, 'maxphase': maxphase, 'phase1': phase1,
            'phase2': phase2, 'trmodel1': trmodel1, 'trmodel2': trmodel2, 'eventType': eventType, 'planet period':period}


def planet_spec(planet, star, w_unit, disperser, deptherr, nchan, smooth=None):
    '''Plot exoplanet transmission/emission spectrum

    Parameters
    ----------
    planet: dict
        planet dictionary from exo_input
    star : dict
        star dictionary from exo_input
    w_unit : str
        wavelength unit (um or nm)
    disperser :
        grism (g102 or g141)
    deptherr : float
        simulated transit/eclipse depth uncertainty
    nchan : float
        number of spectrophotometric channels
    smooth : float
        (Optional)length of smoothing kernel

     Returns
     -------
     dict
        contains following keys {'model_wave','model_spec','binwave','binspec',
        'error','wmin','wmax'}
     '''
    # Load model wavelengths and spectrum
    mwave, mspec = hst_spec(planet, star)  # np.loadtxt(specfile, unpack=True)
    # Convert wavelength to microns
    # if w_unit == 'um':
    #    pass
    # elif w_unit == 'nm':
    #    mwave /= 1000.
    # else:
    #    print(("****HALTED: Unrecognized wavelength unit: '%s'" % w_unit))
    #    return

    # Smooth model spectrum (optional)
    if smooth != None:
        from .hst_smooth import smooth as sm
        mspec = sm(mspec, smooth)

    # Determine disperser wavelength boundaries
    if disperser == 'g141':
        wmin = 1.125
        wmax = 1.650
    elif disperser == 'g102':
        wmin = 0.84
        wmax = 1.13
    else:
        print(("****HALTED: Unrecognized disperser name: '%s'" % disperser))
        return

    # Determine wavelength bins
    binsize = (wmax - wmin)/nchan
    wave_low = np.round([i for i in np.linspace(wmin, wmax-binsize, nchan)], 3)
    wave_hi = np.round([i for i in np.linspace(wmin+binsize, wmax, nchan)], 3)
    binwave = (wave_low + wave_hi)/2.

    # Create simulated spectrum by binning model spectrum and addding uncertainty
    binspec = np.zeros(nchan)
    for i in range(nchan):
        ispec = np.where((mwave >= wave_low[i])*(mwave <= wave_hi[i]))
        binspec[i] = np.mean(mspec[ispec])
    binspec += np.random.normal(0, deptherr, nchan)

    return {'model_wave': mwave, 'model_spec': mspec, 'binwave': binwave, 'binspec': binspec,
            'error': deptherr, 'wmin': wmin, 'wmax': wmax}


def compute_sim_lightcurve(exposureDict, lightCurveDict, calRamp=False):
    """Compute simulated HST light curves

    Function to take the fluence and error estimates from wfc3_TExoNS
    and the model light curve from calc_start_window to simulate observed
    light curves. Ramp effect systemacts simulated by RECTE can be
    included by turning on the calRamp switch

    Parameters
    ----------
    exposureDict : dict
    Includes information for the observation. This dictionary is
    returned by function wfc3_TExoNS. Relevant keys in the dictionary
    are ['info']["Maximum pixel fluence (electrons)"] and
    ['info']['exposure time']

    lightCurveDict : dict
    Includes the model light curve. The light curves are estimed by
    calc_start_window

    calRamp : bool (default: False)
    Switch to turn on/off RECTE ramp calculation. Calculate realistic
    ramp effect systematics when the switch is turned on.

    Returns
    -------
    dict
    Resulting light curve (unit: e/pixel) for the earliest and latest time

    """
    fluence = exposureDict['info']["Maximum pixel fluence (electrons)"]
    exptime = exposureDict['info']['exposure time']
    obst1 = (lightCurveDict['obsphase1'] -
          lightCurveDict['obsphase1'][0]) *\
          lightCurveDict['planet period'] * 86400 # in seconds
    obst2 = (lightCurveDict['obsphase2'] -
          lightCurveDict['obsphase2'][0]) *\
          lightCurveDict['planet period'] * 86400 # in seconds
    counts1 = lightCurveDict['obstr1'] * fluence
    counts2 = lightCurveDict['obstr2'] * fluence
    # use RECTE to calculate the ramp if calRamp option is turned on
    if calRamp:
        counts1 = RECTE(counts1 / exptime,
                       obst1,
                       exptime)
        counts2 = RECTE(counts2 / exptime,
                       obst2,
                       exptime)
    count_noise = lightCurveDict['light_curve_rms'] * fluence
    resultDict = lightCurveDict.copy()
    resultDict['counts1'] = counts1
    resultDict['counts2'] = counts2
    resultDict['count_noise'] = count_noise
    resultDict['ramp_included'] = calRamp
    resultDict['model_counts1'] = resultDict['trmodel1'] * fluence
    resultDict['model_counts2'] = resultDict['trmodel2'] * fluence
    return resultDict


def compute_sim_hst(dictinput,verbose=False):
    """Sets up HST simulations

    Function to set up explanet observations for HST only and
    compute simulated spectrum.

    Parameters
    ----------
    dictinput : dict
        instrument and pandexo dictionaries in format {"pandeia_input":dict1, "pandexo_input":dict2}

    Returns
    -------
    dict
        All hst output info needed to plot simulated data, light curves timing info
    """
    pandexo_input = dictinput['pandexo_input']
    pandeia_input = dictinput['pandeia_input']

    disperser = pandeia_input['configuration']['instrument']['disperser'].lower()

    # add a switch for ramp calculation
    calRamp = pandeia_input['strategy']['calculateRamp']

    if not pandeia_input['strategy']['useFirstOrbit']:
        if verbose:print("Dropping first orbit designed by observation strategy")
        if verbose:print("Do not calculate ramp profile")
        calRamp = False
    numorbits = pandeia_input['strategy']['norbits']
    nchan = pandeia_input['strategy']['nchan']
    windowSize = pandeia_input['strategy']['windowSize']
    useFirstOrbit = pandeia_input['strategy']['useFirstOrbit']
    calc_type = pandexo_input['planet']['type']
    jmag = pandexo_input['star']['jmag']
    hmag = pandexo_input['star']['hmag']
    specfile = pandexo_input['planet']['exopath']
    w_unit = pandexo_input['planet']['w_unit']
    f_unit = pandexo_input['planet']['f_unit']
    depth = pandexo_input['planet']['depth']
    inc = pandexo_input['planet']['i']
    aRs = pandexo_input['planet']['ars']
    period = pandexo_input['planet']['period']
    ecc = pandexo_input['planet']['ecc']
    w = pandexo_input['planet']['w']
    try:
        offset = pandeia_input['strategy']['offset']
    except:
        offset = 0.
    
    # check to see if ecc or w was provided
    if (type(ecc) != float) and (type(ecc) != int):
        ecc = 0.0
    if (type(w) != float) and (type(w) != int):
        w = 90.0
    if (type(windowSize) != float) and (type(windowSize) != int):
        windowSize = 20.0
    if f_unit == "rp^2/r*^2":
        eventType = 'transit'
    elif f_unit == "fp/f*":
        eventType = 'eclipse'
    elif calc_type == 'grid':
        eventType = 'transit'
    else:
        raise Exception('Units are not correct. Pick rp^2/r*^2 or fp/f*')

    a = wfc3_TExoNS(dictinput)

    b = calc_start_window(eventType, a['light_curve_rms'], a['nframes_per_orb'], a['info']['Number of HST orbits'],
        depth, inc, aRs, period, windowSize, ecc, w, useFirstOrbit=useFirstOrbit, offset=offset)
    c = planet_spec(pandexo_input['planet'], pandexo_input['star'],
                    w_unit, disperser, a['spec_error'], nchan, smooth=20)
    info_div = create_out_div(a['info'], b['minphase'], b['maxphase'])
    simLightCurve = compute_sim_lightcurve(a, b, calRamp=calRamp)
    return {"wfc3_TExoNS": a,
            "calc_start_window": b,
            "planet_spec": c,
            "light_curve": simLightCurve,
            "info_div": info_div}


def create_out_div(input_dict, minphase, maxphase):
    """Function to render input dicts in html format for web front end

    Parameters
    ----------
    input_dict : dict
        any input dictionary

    Returns
    -------
    div
        html rendered table
    """
    input_dict['Start observations between orbital phases'] = str(
        minphase)+'-'+str(maxphase)
    input_div = pd.DataFrame.from_dict(input_dict, orient='index')
    input_div.columns = ['Value']
    input_div = input_div.to_html()
    input_div = '<table class="table table-striped"> \n' + \
        input_div[36:len(input_div)]
    input_div = input_div.encode()
    return input_div

def calculate_tsec(period, ecc, omega, inc, t0 = None, tperi = None, winn_approximation = False):
    ''' Function to calculate the time of secondary eclipse.

        This uses Halley's method (Newton-Raphson, but using second derivatives) to first find the true anomaly (f) at which secondary eclipse occurs,
        then uses this to get the eccentric anomaly (E) at secondary eclipse, which gives the mean anomaly (M) at secondary
        eclipse using Kepler's equation. This finally leads to the time of secondary eclipse using the definition of the mean
        anomaly (M = n*(t - tau) --- here tau is the time of pericenter passage, n = 2*pi/period the mean motion).
        Time inputs can be either the time of periastron passage directly or the time of transit center. If the latter, the
        true anomaly for primary transit will be calculated using Halley's method as well, and this will be used to get the
        time of periastron passage.

        Parameters
        ----------
        period : float
            The period of the transit in days.
        ecc : float
            Eccentricity of the orbit
        omega : float
            Argument of periastron passage (in radians)
        inc : string
            Inclination of the orbit (in radians)
        t0 : float
            The transit time in BJD or HJD (will be used to get time of periastron passage).
        tperi : float
            The time of periastron passage in BJD or HJD (needed if t0 is not supplied).
        winn_approximation : boolean
            If True, the approximation in Winn (2010) is used --- (only valid for not very eccentric and inclined orbits).
        Returns
        -------
        tsec : float
            The time of secondary eclipse '''
    # Check user is not playing trick on us:
    if period<0:
        raise Exception('Period cannot be a negative number.')
    if ecc<0 or ecc>1:
        raise Exception('Eccentricity (e) is out of bounds (0 < e < 1).')

    # Use true anomaly approximation given in Winn (2010) as starting point:
    f_occ_0 = (-0.5*np.pi) - omega
    if not winn_approximation:
        f_occ = optimize.newton(drsky, f_occ_0, fprime = drsky_prime, fprime2 = drsky_2prime, args = (ecc, omega, inc,))
    else:
        f_occ = f_occ_0
    # Define the mean motion, n:
    n = 2.*np.pi/period

    # If time of transit center is given, use it to calculate the time of periastron passage. If no time of periastron
    # or time-of-transit center given, raise error:
    if tperi is None:
        # For this, find true anomaly during transit. Use Winn (2010) as starting point:
        f_tra_0 = (np.pi/2.) - omega
        if not winn_approximation:
            f_tra = optimize.newton(drsky, f_tra_0, fprime = drsky_prime, fprime2 = drsky_2prime, args = (ecc, omega, inc,))
        else:
            f_tra = f_tra_0
        # Get eccentric anomaly during transit:
        E = getE(f_tra, ecc)

        # Get mean anomaly during transit:
        M = getM(E, ecc)

        # Get time of periastron passage from mean anomaly definition:
        tperi = t0 - (M/n)

    elif (tperi is None) and (t0 is None):
        raise ValueError('The time of periastron passage or time-of-transit center has to be supplied for the calculation to work.')

    # Get eccentric anomaly:
    E = getE(f_occ, ecc)

    # Get mean anomaly during secondary eclipse:
    M = getM(E, ecc)

    # Get the time of secondary eclipse using the definition of the mean anomaly:
    tsec = (M/n) + tperi

    # Note returned time-of-secondary eclipse is the closest to the time of periastron passage and/or time-of-transit center. Check that
    # the returned tsec is the *next* tsec to the time of periastron or t0 (i.e., the closest *future* tsec):
    if t0 is not None:
        tref = t0
    else:
        tref = tperi
    if tref > tsec:
        while True:
            tsec += period
            if tsec > tref:
                break
    return tsec

def drsky_2prime(x, ecc, omega, inc):
    ''' Second derivative of function drsky. This is the second derivative with respect to f of the drsky function.
    Parameters
    ----------
    x : float
      True anomaly
    ecc : float
      Eccentricity of the orbit
    omega : float
      Argument of periastron passage (in radians)
    inc : float
      Inclination of the orbit (in radians)

    Returns
    -------
    drsky_2prime : float
      Function evaluated at x, ecc, omega, inc'''

    sq_sini = np.sin(inc)**2
    sin_o_p_f = np.sin(x+omega)
    cos_o_p_f = np.cos(x+omega)
    ecosf = ecc*np.cos(x)
    esinf = ecc*np.sin(x)

    f1 = esinf - esinf*sq_sini*(sin_o_p_f**2)
    f2 = -sq_sini*(ecosf + 4.)*(sin_o_p_f*cos_o_p_f)

    return f1+f2

def drsky_prime(x, ecc, omega, inc):
    ''' Derivative of function drsky. This is the first derivative with respect to f of the drsky function.
    Parameters
    ----------
    x : float
      True anomaly
    ecc : float
      Eccentricity of the orbit
    omega : float
      Argument of periastron passage (in radians)
    inc : float
      Inclination of the orbit (in radians)

    Returns
    -------
    drsky_prime : float
      Function evaluated at x, ecc, omega, inc'''

    sq_sini = np.sin(inc)**2
    sin_o_p_f = np.sin(x+omega)
    cos_o_p_f = np.cos(x+omega)
    ecosf = ecc*np.cos(x)
    esinf = ecc*np.sin(x)

    f1 = (cos_o_p_f**2 - sin_o_p_f**2)*(sq_sini)*(1. + ecosf)
    f2 = -ecosf*(1 - (sin_o_p_f**2)*(sq_sini))
    f3 = esinf*sin_o_p_f*cos_o_p_f*sq_sini

    return f1+f2+f3

def drsky(x, ecc, omega, inc):
    ''' Function whose roots we wish to find to obtain time of secondary (and primary) eclipse(s)
    When one takes the derivative of equation (5) in Winn (2010; https://arxiv.org/abs/1001.2010v5), and equates that to zero (to find the
    minimum/maximum of said function), one gets to an equation of the form g(x) = 0. This function (drsky) is g(x), where x is the true
    anomaly.
    Parameters
    ----------
    x : float
      True anomaly
    ecc : float
      Eccentricity of the orbit
    omega : float
      Argument of periastron passage (in radians)
    inc : float
      Inclination of the orbit (in radians)

    Returns
    -------
    drsky : float
      Function evaluated at x, ecc, omega, inc '''

    sq_sini = np.sin(inc)**2
    sin_o_p_f = np.sin(x+omega)
    cos_o_p_f = np.cos(x+omega)

    f1 = sin_o_p_f*cos_o_p_f*sq_sini*(1. + ecc*np.cos(x))
    f2 = ecc*np.sin(x)*(1. - sin_o_p_f**2 * sq_sini)
    return f1 - f2

def getE(f,ecc):
    """ Function that returns the eccentric anomaly
    Note normally this is defined in terms of cosines (see, e.g., Section 2.4 in Murray and Dermott), but numerically
    this is troublesome because the arccosine doesn't handle negative numbers by definition (equation 2.43). That's why
    the arctan version is better as signs are preserved (derivation is also in the same section, equation 2.46).
    Parameters
    ----------
    f : float
      True anomaly
    ecc : float
      Eccentricity
    Returns
    -------
    E : float
      Eccentric anomaly """

    return 2. * np.arctan(np.sqrt((1.-ecc)/(1.+ecc))*np.tan(f/2.))

def getM(E, ecc):
    """ Function that returns the mean anomaly using Kepler's equation
    Parameters
    ----------
    E : float
      Eccentric anomaly
    ecc: float
      Eccentricity
    Returns
    -------
    M : float
      Mean anomaly """

    return E - ecc*np.sin(E)
